# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

import six

from glance.common import exception
from glance.common.glare import declarative
from glance.common.glare import definitions
from glance import glare as ga
from glance.i18n import _


COMMON_ARTIFACT_PROPERTIES = ['id',
                              'type_name',
                              'type_version',
                              'name',
                              'version',
                              'description',
                              'visibility',
                              'state',
                              'tags',
                              'owner',
                              'created_at',
                              'updated_at',
                              'published_at',
                              'deleted_at']


def _serialize_list_prop(prop, values):
    """
    A helper func called to correctly serialize an Array property.

    Returns a dict {'type': some_supported_db_type, 'value': serialized_data}
    """
    # FIXME(Due to a potential bug in declarative framework, for Arrays, that
    # are values to some dict items (Dict(properties={"foo": Array()})),
    # prop.get_value(artifact) returns not the real list of items, but the
    # whole dict). So we can't rely on prop.get_value(artifact) and will pass
    # correctly retrieved values to this function
    serialized_value = []
    for i, val in enumerate(values or []):
        db_type = prop.get_item_definition_at_index(i).DB_TYPE
        if db_type is None:
            continue
        serialized_value.append({
            'type': db_type,
            'value': val
        })
    return serialized_value


def _serialize_dict_prop(artifact, prop, key, value, save_prop_func):
    key_to_save = prop.name + '.' + key
    dict_key_prop = prop.get_prop_definition_at_key(key)
    db_type = dict_key_prop.DB_TYPE
    if (db_type is None and
        not isinstance(dict_key_prop,
                       declarative.ListAttributeDefinition)):
        # nothing to do here, don't know how to deal with this type
        return
    elif isinstance(dict_key_prop,
                    declarative.ListAttributeDefinition):
        serialized = _serialize_list_prop(
            dict_key_prop,
            # FIXME(see comment for _serialize_list_prop func)
            values=(dict_key_prop.get_value(artifact) or {}).get(key, []))
        save_prop_func(key_to_save, 'array', serialized)
    else:
        save_prop_func(key_to_save, db_type, value)


def _serialize_dependencies(artifact):
    """Returns a dict of serialized dependencies for given artifact"""
    dependencies = {}
    for relation in artifact.metadata.attributes.dependencies.values():
        serialized_dependency = []
        if isinstance(relation, declarative.ListAttributeDefinition):
            for dep in relation.get_value(artifact):
                serialized_dependency.append(dep.id)
        else:
            relation_data = relation.get_value(artifact)
            if relation_data:
                serialized_dependency.append(relation.get_value(artifact).id)
        dependencies[relation.name] = serialized_dependency
    return dependencies


def _serialize_blobs(artifact):
    """Return a dict of serialized blobs for given artifact"""
    blobs = {}
    for blob in artifact.metadata.attributes.blobs.values():
        serialized_blob = []
        if isinstance(blob, declarative.ListAttributeDefinition):
            for b in blob.get_value(artifact) or []:
                serialized_blob.append({
                    'size': b.size,
                    'locations': b.locations,
                    'checksum': b.checksum,
                    'item_key': b.item_key
                })
        else:
            b = blob.get_value(artifact)
            # if no value for blob has been set -> continue
            if not b:
                continue
            serialized_blob.append({
                'size': b.size,
                'locations': b.locations,
                'checksum': b.checksum,
                'item_key': b.item_key
            })
        blobs[blob.name] = serialized_blob
    return blobs


def serialize_for_db(artifact):
    result = {}
    custom_properties = {}

    def _save_prop(prop_key, prop_type, value):
        custom_properties[prop_key] = {
            'type': prop_type,
            'value': value
        }

    for prop in artifact.metadata.attributes.properties.values():
        if prop.name in COMMON_ARTIFACT_PROPERTIES:
            result[prop.name] = prop.get_value(artifact)
            continue
        if isinstance(prop, declarative.ListAttributeDefinition):
            serialized_value = _serialize_list_prop(prop,
                                                    prop.get_value(artifact))
            _save_prop(prop.name, 'array', serialized_value)
        elif isinstance(prop, declarative.DictAttributeDefinition):
            fields_to_set = prop.get_value(artifact) or {}
            # if some keys are not present (like in prop == {}), then have to
            # set their values to None.
            # XXX FIXME prop.properties may be a dict ({'foo': '', 'bar': ''})
            # or String\Integer\whatsoever, limiting the possible dict values.
            # In the latter case have no idea how to remove old values during
            # serialization process.
            if isinstance(prop.properties, dict):
                for key in [k for k in prop.properties
                            if k not in fields_to_set.keys()]:
                    _serialize_dict_prop(artifact, prop, key, None, _save_prop)
            # serialize values of properties present
            for key, value in six.iteritems(fields_to_set):
                _serialize_dict_prop(artifact, prop, key, value, _save_prop)
        elif prop.DB_TYPE is not None:
            _save_prop(prop.name, prop.DB_TYPE, prop.get_value(artifact))

    result['properties'] = custom_properties
    result['dependencies'] = _serialize_dependencies(artifact)
    result['blobs'] = _serialize_blobs(artifact)
    return result


def _deserialize_blobs(artifact_type, blobs_from_db, artifact_properties):
    """Retrieves blobs from database"""
    for blob_name, blob_value in six.iteritems(blobs_from_db):
        if not blob_value:
            continue
        if isinstance(artifact_type.metadata.attributes.blobs.get(blob_name),
                      declarative.ListAttributeDefinition):
            val = []
            for v in blob_value:
                b = definitions.Blob(size=v['size'],
                                     locations=v['locations'],
                                     checksum=v['checksum'],
                                     item_key=v['item_key'])
                val.append(b)
        elif len(blob_value) == 1:
            val = definitions.Blob(size=blob_value[0]['size'],
                                   locations=blob_value[0]['locations'],
                                   checksum=blob_value[0]['checksum'],
                                   item_key=blob_value[0]['item_key'])
        else:
            raise exception.InvalidArtifactPropertyValue(
                message=_('Blob %(name)s may not have multiple values'),
                name=blob_name)
        artifact_properties[blob_name] = val


def _deserialize_dependencies(artifact_type, deps_from_db,
                              artifact_properties, plugins):
    """Retrieves dependencies from database"""
    for dep_name, dep_value in six.iteritems(deps_from_db):
        if not dep_value:
            continue
        if isinstance(
                artifact_type.metadata.attributes.dependencies.get(dep_name),
                declarative.ListAttributeDefinition):
            val = []
            for v in dep_value:
                val.append(deserialize_from_db(v, plugins))
        elif len(dep_value) == 1:
            val = deserialize_from_db(dep_value[0], plugins)
        else:
            raise exception.InvalidArtifactPropertyValue(
                message=_('Relation %(name)s may not have multiple values'),
                name=dep_name)
        artifact_properties[dep_name] = val


def deserialize_from_db(db_dict, plugins):
    artifact_properties = {}
    type_name = None
    type_version = None

    for prop_name in COMMON_ARTIFACT_PROPERTIES:
        prop_value = db_dict.pop(prop_name, None)
        if prop_name == 'type_name':
            type_name = prop_value
        elif prop_name == 'type_version':
            type_version = prop_value
        else:
            artifact_properties[prop_name] = prop_value

    try:
        artifact_type = plugins.get_class_by_typename(type_name, type_version)
    except exception.ArtifactPluginNotFound:
        raise exception.UnknownArtifactType(name=type_name,
                                            version=type_version)

    type_specific_properties = db_dict.pop('properties', {})
    for prop_name, prop_value in six.iteritems(type_specific_properties):
        prop_type = prop_value.get('type')
        prop_value = prop_value.get('value')
        if prop_value is None:
            continue
        if '.' in prop_name:  # dict-based property
            name, key = prop_name.split('.', 1)
            artifact_properties.setdefault(name, {})
            if prop_type == 'array':
                artifact_properties[name][key] = [item.get('value') for item in
                                                  prop_value]
            else:
                artifact_properties[name][key] = prop_value
        elif prop_type == 'array':  # list-based property
            artifact_properties[prop_name] = [item.get('value') for item in
                                              prop_value]
        else:
            artifact_properties[prop_name] = prop_value

    blobs = db_dict.pop('blobs', {})
    _deserialize_blobs(artifact_type, blobs, artifact_properties)

    dependencies = db_dict.pop('dependencies', {})
    _deserialize_dependencies(artifact_type, dependencies,
                              artifact_properties, plugins)

    return artifact_type(**artifact_properties)


def _process_blobs_for_client(artifact, result):
    """Processes artifact's blobs: adds download links and pretty-printed data.

    The result is stored in 'result' dict.
    """
    def build_uri(blob_attr, position=None):
        """A helper func to build download uri"""
        template = "/artifacts/%(type)s/v%(version)s/%(id)s/%(prop)s/download"
        format_dict = {
            "type": artifact.metadata.endpoint,
            "version": artifact.type_version,
            "id": artifact.id,
            "prop": blob_attr.name
        }
        if position is not None:
            template = ("/artifacts/%(type)s/v%(version)s/"
                        "%(id)s/%(prop)s/%(position)s/download")
            format_dict["position"] = position

        return template % format_dict

    for blob_attr in artifact.metadata.attributes.blobs.values():
        value = blob_attr.get_value(artifact)
        if value is None:
            result[blob_attr.name] = None
        elif isinstance(value, collections.Iterable):
            res_list = []
            for pos, blob in enumerate(value):
                blob_dict = blob.to_dict()
                blob_dict["download_link"] = build_uri(blob_attr, pos)
                res_list.append(blob_dict)
            result[blob_attr.name] = res_list
        else:
            result[blob_attr.name] = value.to_dict()
            result[blob_attr.name]["download_link"] = build_uri(blob_attr)


def serialize_for_client(artifact, show_level=ga.Showlevel.NONE):
    # use serialize_for_db and modify some fields
    # (like properties, show only value, not type)
    result = {}

    for prop in artifact.metadata.attributes.properties.values():
        result[prop.name] = prop.get_value(artifact)

    if show_level > ga.Showlevel.NONE:
        for dep in artifact.metadata.attributes.dependencies.values():
            inner_show_level = (ga.Showlevel.DIRECT
                                if show_level == ga.Showlevel.DIRECT
                                else ga.Showlevel.NONE)
            value = dep.get_value(artifact)
            if value is None:
                result[dep.name] = None
            elif isinstance(value, list):
                result[dep.name] = [serialize_for_client(v, inner_show_level)
                                    for v in value]
            else:
                result[dep.name] = serialize_for_client(value,
                                                        inner_show_level)
    _process_blobs_for_client(artifact, result)
    return result



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import mock
import pkg_resources

from glance.common import exception
from glance.common.glare import loader
from glance.contrib.plugins.artifacts_sample.v1 import artifact as art1
from glance.contrib.plugins.artifacts_sample.v2 import artifact as art2
from glance.tests import utils


class MyArtifactDuplicate(art1.MyArtifact):
    __type_version__ = '1.0.1'
    __type_name__ = 'MyArtifact'


class MyArtifactOk(art1.MyArtifact):
    __type_version__ = '1.0.2'
    __type_name__ = 'MyArtifact'


class TestArtifactsLoader(utils.BaseTestCase):
    def setUp(self):
        self.path = 'glance.contrib.plugins.artifacts_sample'
        self._setup_loader(['MyArtifact=%s.v1.artifact:MyArtifact' %
                            self.path])
        super(TestArtifactsLoader, self).setUp()

    def _setup_loader(self, artifacts):
        self.loader = None
        mock_this = 'stevedore.extension.ExtensionManager._find_entry_points'
        with mock.patch(mock_this) as fep:
            fep.return_value = [
                pkg_resources.EntryPoint.parse(art) for art in artifacts]
            self.loader = loader.ArtifactsPluginLoader(
                'glance.artifacts.types')

    def test_load(self):
        """
        Plugins can be loaded as entrypoint=single plugin and
        entrypoint=[a, list, of, plugins]
        """
        # single version
        self.assertEqual(1, len(self.loader.mgr.extensions))
        self.assertEqual(art1.MyArtifact,
                         self.loader.get_class_by_endpoint('myartifact'))
        # entrypoint = [a, list]
        path = os.path.splitext(__file__)[0][__file__.rfind(
            'glance'):].replace('/', '.')
        self._setup_loader([
            'MyArtifact=%s:MyArtifactOk' % path,
            'MyArtifact=%s.v2.artifact:MyArtifact' % self.path,
            'MyArtifact=%s.v1.artifact:MyArtifact' % self.path]),
        self.assertEqual(3, len(self.loader.mgr.extensions))
        # returns the plugin with the latest version
        self.assertEqual(art2.MyArtifact,
                         self.loader.get_class_by_endpoint('myartifact'))
        self.assertEqual(art1.MyArtifact,
                         self.loader.get_class_by_endpoint('myartifact',
                                                           '1.0.1'))

    def test_basic_loader_func(self):
        """Test public methods of PluginLoader class here"""
        # type_version 2 == 2.0 == 2.0.0
        self._setup_loader(
            ['MyArtifact=%s.v2.artifact:MyArtifact' % self.path])
        self.assertEqual(art2.MyArtifact,
                         self.loader.get_class_by_endpoint('myartifact'))
        self.assertEqual(art2.MyArtifact,
                         self.loader.get_class_by_endpoint('myartifact',
                                                           '2.0'))
        self.assertEqual(art2.MyArtifact,
                         self.loader.get_class_by_endpoint('myartifact',
                                                           '2.0.0'))
        self.assertEqual(art2.MyArtifact,
                         self.loader.get_class_by_endpoint('myartifact',
                                                           '2'))
        # now make sure that get_class_by_typename works as well
        self.assertEqual(art2.MyArtifact,
                         self.loader.get_class_by_typename('MyArtifact'))
        self.assertEqual(art2.MyArtifact,
                         self.loader.get_class_by_typename('MyArtifact', '2'))

    def test_config_validation(self):
        """
        Plugins can be loaded on certain conditions:
            * entry point name == type_name
            * no plugin with the same type_name and version has been already
              loaded
        """
        path = 'glance.contrib.plugins.artifacts_sample'
        # here artifacts specific validation is checked
        self.assertRaises(exception.ArtifactNonMatchingTypeName,
                          self._setup_loader,
                          ['non_matching_name=%s.v1.artifact:MyArtifact' %
                           path])
        # make sure this call is ok
        self._setup_loader(['MyArtifact=%s.v1.artifact:MyArtifact' % path])
        art_type = self.loader.get_class_by_endpoint('myartifact')
        self.assertEqual('MyArtifact', art_type.metadata.type_name)
        self.assertEqual('1.0.1', art_type.metadata.type_version)
        # now try to add duplicate artifact with the same type_name and
        # type_version as already exists
        bad_art_path = os.path.splitext(__file__)[0][__file__.rfind(
            'glance'):].replace('/', '.')
        self.assertEqual(art_type.metadata.type_version,
                         MyArtifactDuplicate.metadata.type_version)
        self.assertEqual(art_type.metadata.type_name,
                         MyArtifactDuplicate.metadata.type_name)
        # should raise an exception as (name, version) is not unique
        self.assertRaises(
            exception.ArtifactDuplicateNameTypeVersion, self._setup_loader,
            ['MyArtifact=%s.v1.artifact:MyArtifact' % path,
             'MyArtifact=%s:MyArtifactDuplicate' % bad_art_path])
        # two artifacts with the same name but different versions coexist fine
        self.assertEqual('MyArtifact', MyArtifactOk.metadata.type_name)
        self.assertNotEqual(art_type.metadata.type_version,
                            MyArtifactOk.metadata.type_version)
        self._setup_loader(['MyArtifact=%s.v1.artifact:MyArtifact' % path,
                            'MyArtifact=%s:MyArtifactOk' % bad_art_path])

    def test_check_function(self):
        """
        A test to show that plugin-load specific options in artifacts.conf
        are correctly processed:

        * no plugins can be loaded if load_enabled = False
        * if available_plugins list is given only plugins specified can be
          be loaded

        """
        self.config(load_enabled=False)
        self.assertRaises(exception.ArtifactLoadError,
                          self._setup_loader,
                          ['MyArtifact=%s.v1.artifact:MyArtifact' % self.path])
        self.config(load_enabled=True, available_plugins=['MyArtifact-1.0.2'])
        self.assertRaises(exception.ArtifactLoadError,
                          self._setup_loader,
                          ['MyArtifact=%s.v1.artifact:MyArtifact' % self.path])
        path = os.path.splitext(__file__)[0][__file__.rfind(
            'glance'):].replace('/', '.')
        self._setup_loader(['MyArtifact=%s:MyArtifactOk' % path])
        # make sure that plugin_map has the expected plugin
        self.assertEqual(MyArtifactOk,
                         self.loader.get_class_by_endpoint('myartifact',
                                                           '1.0.2'))



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys

import glance_store
import jsonschema
from oslo_config import cfg
from oslo_serialization import jsonutils as json
from oslo_utils import encodeutils
from oslo_utils import excutils
import semantic_version
import six
import six.moves.urllib.parse as urlparse
import webob.exc

from glance.common import exception
from glance.common.glare import loader
from glance.common.glare import serialization
from glance.common import jsonpatchvalidator
from glance.common import utils
from glance.common import wsgi
import glance.db
from glance.glare import gateway
from glance.glare import Showlevel
from glance.i18n import _, _LE
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

possible_topdir = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
                                                os.pardir,
                                                os.pardir))
if os.path.exists(os.path.join(possible_topdir, 'glance', '__init__.py')):
    sys.path.insert(0, possible_topdir)

CONF = cfg.CONF
CONF.import_group("profiler", "glance.common.wsgi")


class ArtifactsController(object):
    def __init__(self, db_api=None, store_api=None, plugins=None):
        self.db_api = db_api or glance.db.get_api()
        self.store_api = store_api or glance_store
        self.plugins = plugins or loader.ArtifactsPluginLoader(
            'glance.artifacts.types')
        self.gateway = gateway.Gateway(self.db_api,
                                       self.store_api, self.plugins)

    @staticmethod
    def _do_update_op(artifact, change):
        """Call corresponding method of the updater proxy.

        Here 'change' is a typical jsonpatch request dict:
            * 'path' - a json-pointer string;
            * 'op' - one of the allowed operation types;
            * 'value' - value to set (omitted when op = remove)
        """
        update_op = getattr(artifact, change['op'])
        update_op(change['path'], change.get('value'))
        return artifact

    @staticmethod
    def _get_artifact_with_dependencies(repo, art_id,
                                        type_name=None, type_version=None):
        """Retrieves an artifact with dependencies from db by its id.

        Show level is direct (only direct dependencies are shown).
        """
        return repo.get(art_id, show_level=Showlevel.DIRECT,
                        type_name=type_name, type_version=type_version)

    def show(self, req, type_name, type_version,
             show_level=Showlevel.TRANSITIVE, **kwargs):
        """Retrieves one artifact by id with its dependencies"""
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        try:
            art_id = kwargs.get('id')
            artifact = artifact_repo.get(art_id, type_name=type_name,
                                         type_version=type_version,
                                         show_level=show_level)
            return artifact
        except exception.ArtifactNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)

    def list(self, req, type_name, type_version, state, **kwargs):
        """Retrieves a list of artifacts that match some params"""
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        filters = kwargs.pop('filters', {})

        filters.update(type_name={'value': type_name},
                       state={'value': state})
        if type_version is not None:
            filters['type_version'] = {'value': type_version}
        if 'version' in filters:
            for filter in filters['version']:
                if filter['value'] == 'latest':
                    if 'name' not in filters:
                        raise webob.exc.HTTPBadRequest(
                            'Filtering by latest version without specifying'
                            ' a name is not supported.')
                    filter['value'] = self._get_latest_version(
                        req, filters['name'][0]['value'], type_name,
                        type_version)
                else:
                    try:
                        semantic_version.Version(filter['value'], partial=True)
                    except ValueError:
                        msg = (_('The format of the version %s is not valid. '
                                 'Use semver notation') % filter['value'])
                        raise webob.exc.HTTPBadRequest(explanation=msg)

        res = artifact_repo.list(filters=filters,
                                 show_level=Showlevel.BASIC,
                                 **kwargs)
        result = {'artifacts': res}
        limit = kwargs.get("limit")
        if limit is not None and len(res) != 0 and len(res) == limit:
            result['next_marker'] = res[-1].id
        return result

    def _get_latest_version(self, req, name, type_name, type_version=None,
                            state='creating'):
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        filters = dict(name=[{"value": name}],
                       type_name={"value": type_name},
                       state={"value": state})
        if type_version is not None:
            filters["type_version"] = {"value": type_version}
        result = artifact_repo.list(filters=filters,
                                    show_level=Showlevel.NONE,
                                    sort_keys=[('version', None)])
        if len(result):
            return result[0].version

        msg = "No artifacts have been found"
        raise exception.ArtifactNotFound(message=msg)

    @utils.mutating
    def create(self, req, artifact_type, artifact_data, **kwargs):
        try:
            artifact_factory = self.gateway.get_artifact_type_factory(
                req.context, artifact_type)
            new_artifact = artifact_factory.new_artifact(**artifact_data)
            artifact_repo = self.gateway.get_artifact_repo(req.context)
            artifact_repo.add(new_artifact)
            # retrieve artifact from db
            return self._get_artifact_with_dependencies(artifact_repo,
                                                        new_artifact.id)
        except (TypeError,
                exception.ArtifactNotFound,
                exception.Invalid,
                exception.DuplicateLocation) as e:
            raise webob.exc.HTTPBadRequest(explanation=e)
        except exception.Forbidden as e:
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.LimitExceeded as e:
            raise webob.exc.HTTPRequestEntityTooLarge(
                explanation=e.msg, request=req, content_type='text/plain')
        except exception.Duplicate as e:
            raise webob.exc.HTTPConflict(explanation=e.msg)
        except exception.NotAuthenticated as e:
            raise webob.exc.HTTPUnauthorized(explanation=e.msg)

    @utils.mutating
    def update_property(self, req, id, type_name, type_version, path, data,
                        **kwargs):
        """Updates a single property specified by request url."""
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        try:
            artifact = self._get_artifact_with_dependencies(artifact_repo, id,
                                                            type_name,
                                                            type_version)
            self._ensure_write_access(artifact, req.context)
            if artifact.metadata.attributes.blobs.get(path) is not None:
                msg = _('Invalid Content-Type for work with %s') % path
                raise webob.exc.HTTPBadRequest(explanation=msg)

            # use updater mixin to perform updates: generate update path
            if req.method == "PUT":
                # replaces existing value or creates a new one
                if getattr(artifact, kwargs["attr"]):
                    artifact.replace(path=path, value=data)
                else:
                    artifact.add(path=path, value=data)
            else:
                # append to an existing value or create a new one
                artifact.add(path=path, value=data)
            artifact_repo.save(artifact)
            return self._get_artifact_with_dependencies(artifact_repo, id)
        except (exception.InvalidArtifactPropertyValue,
                exception.ArtifactInvalidProperty,
                exception.InvalidJsonPatchPath,
                exception.ArtifactCircularDependency) as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        except exception.NotAuthenticated as e:
            raise webob.exc.HTTPUnauthorized(explanation=e.msg)

    @utils.mutating
    def update(self, req, id, type_name, type_version, changes, **kwargs):
        """Performs an update via json patch request"""
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        try:
            artifact = self._get_artifact_with_dependencies(artifact_repo, id,
                                                            type_name,
                                                            type_version)
            self._ensure_write_access(artifact, req.context)
            updated = artifact
            for change in changes:
                if artifact.metadata.attributes.blobs.get(change['path']):
                    msg = _('Invalid request PATCH for work with blob')
                    raise webob.exc.HTTPBadRequest(explanation=msg)
                else:
                    updated = self._do_update_op(updated, change)
            artifact_repo.save(updated)
            return self._get_artifact_with_dependencies(artifact_repo, id)
        except (exception.InvalidJsonPatchPath,
                exception.Invalid) as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        except exception.Forbidden as e:
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.StorageQuotaFull as e:
            msg = (_("Denying attempt to upload artifact because it exceeds "
                     "the quota: %s") % encodeutils.exception_to_unicode(e))
            raise webob.exc.HTTPRequestEntityTooLarge(
                explanation=msg, request=req, content_type='text/plain')
        except exception.LimitExceeded as e:
            raise webob.exc.HTTPRequestEntityTooLarge(
                explanation=e.msg, request=req, content_type='text/plain')
        except exception.NotAuthenticated as e:
            raise webob.exc.HTTPUnauthorized(explanation=e.msg)

    @utils.mutating
    def delete(self, req, id, type_name, type_version, **kwargs):
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        try:
            artifact = self._get_artifact_with_dependencies(
                artifact_repo, id, type_name=type_name,
                type_version=type_version)
            self._ensure_write_access(artifact, req.context)
            artifact_repo.remove(artifact)
        except exception.Invalid as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)
        except (glance_store.Forbidden, exception.Forbidden) as e:
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except (glance_store.NotFound, exception.NotFound) as e:
            msg = (_("Failed to find artifact %(artifact_id)s to delete") %
                   {'artifact_id': id})
            raise webob.exc.HTTPNotFound(explanation=msg)
        except glance_store.exceptions.InUseByStore as e:
            msg = (_("Artifact %s could not be deleted "
                     "because it is in use: %s") % (id, e.msg))  # noqa
            raise webob.exc.HTTPConflict(explanation=msg)
        except exception.NotAuthenticated as e:
            raise webob.exc.HTTPUnauthorized(explanation=e.msg)

    @utils.mutating
    def publish(self, req, id, type_name, type_version, **kwargs):
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        try:
            artifact = self._get_artifact_with_dependencies(
                artifact_repo, id, type_name=type_name,
                type_version=type_version)
            self._ensure_write_access(artifact, req.context)
            return artifact_repo.publish(artifact, context=req.context)
        except exception.Forbidden as e:
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        except exception.Invalid as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)
        except exception.NotAuthenticated as e:
            raise webob.exc.HTTPUnauthorized(explanation=e.msg)

    def _upload_list_property(self, method, blob_list, index, data, size):
        if method == 'PUT' and not index and len(blob_list) > 0:
            # PUT replaces everything, so PUT to non-empty collection is
            # forbidden
            raise webob.exc.HTTPMethodNotAllowed(
                explanation=_("Unable to PUT to non-empty collection"))
        if index is not None and index > len(blob_list):
            raise webob.exc.HTTPBadRequest(
                explanation=_("Index is out of range"))
        if index is None:
            # both POST and PUT create a new blob list
            blob_list.append((data, size))
        elif method == 'POST':
            blob_list.insert(index, (data, size))
        else:
            blob_list[index] = (data, size)

    @utils.mutating
    def upload(self, req, id, type_name, type_version, attr, size, data,
               index, **kwargs):
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        artifact = None
        try:
            artifact = self._get_artifact_with_dependencies(artifact_repo,
                                                            id,
                                                            type_name,
                                                            type_version)
            self._ensure_write_access(artifact, req.context)
            blob_prop = artifact.metadata.attributes.blobs.get(attr)
            if blob_prop is None:
                raise webob.exc.HTTPBadRequest(
                    explanation=_("Not a blob property '%s'") % attr)
            if isinstance(blob_prop, list):
                blob_list = getattr(artifact, attr)
                self._upload_list_property(req.method, blob_list,
                                           index, data, size)
            else:
                if index is not None:
                    raise webob.exc.HTTPBadRequest(
                        explanation=_("Not a list property '%s'") % attr)
                setattr(artifact, attr, (data, size))
            artifact_repo.save(artifact)
            return artifact

        except ValueError as e:
            exc_message = encodeutils.exception_to_unicode(e)
            LOG.debug("Cannot save data for artifact %(id)s: %(e)s",
                      {'id': id, 'e': exc_message})
            self._restore(artifact_repo, artifact)
            raise webob.exc.HTTPBadRequest(
                explanation=exc_message)

        except glance_store.StoreAddDisabled:
            msg = _("Error in store configuration. Adding artifacts to store "
                    "is disabled.")
            LOG.exception(msg)
            self._restore(artifact_repo, artifact)
            raise webob.exc.HTTPGone(explanation=msg, request=req,
                                     content_type='text/plain')

        except (glance_store.Duplicate,
                exception.InvalidImageStatusTransition) as e:
            LOG.exception(encodeutils.exception_to_unicode(e))
            raise webob.exc.HTTPConflict(explanation=e.msg, request=req)

        except exception.Forbidden as e:
            msg = ("Not allowed to upload data for artifact %s" %
                   id)
            LOG.debug(msg)
            raise webob.exc.HTTPForbidden(explanation=msg, request=req)

        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)

        except glance_store.StorageFull as e:
            msg = _("Artifact storage media "
                    "is full: %s") % encodeutils.exception_to_unicode(e)
            LOG.error(msg)
            self._restore(artifact_repo, artifact)
            raise webob.exc.HTTPRequestEntityTooLarge(explanation=msg,
                                                      request=req)

        except exception.StorageQuotaFull as e:
            msg = _("Artifact exceeds the storage "
                    "quota: %s") % encodeutils.exception_to_unicode(e)
            LOG.error(msg)
            self._restore(artifact_repo, artifact)
            raise webob.exc.HTTPRequestEntityTooLarge(explanation=msg,
                                                      request=req)

        except exception.ImageSizeLimitExceeded as e:
            msg = _("The incoming artifact blob is "
                    "too large: %s") % encodeutils.exception_to_unicode(e)
            LOG.error(msg)
            self._restore(artifact_repo, artifact)
            raise webob.exc.HTTPRequestEntityTooLarge(explanation=msg,
                                                      request=req)

        except glance_store.StorageWriteDenied as e:
            msg = _("Insufficient permissions on artifact "
                    "storage media: %s") % encodeutils.exception_to_unicode(e)
            LOG.error(msg)
            self._restore(artifact_repo, artifact)
            raise webob.exc.HTTPServiceUnavailable(explanation=msg,
                                                   request=req)

        except webob.exc.HTTPGone as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Failed to upload artifact blob data due to"
                              " HTTP error"))

        except webob.exc.HTTPError as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Failed to upload artifact blob data due to HTTP"
                              " error"))
                self._restore(artifact_repo, artifact)

        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Failed to upload artifact blob data due to "
                                  "internal error"))
                self._restore(artifact_repo, artifact)

    def download(self, req, id, type_name, type_version, attr, index,
                 **kwargs):
        artifact_repo = self.gateway.get_artifact_repo(req.context)
        try:
            artifact = artifact_repo.get(id, type_name, type_version)
            if attr in artifact.metadata.attributes.blobs:
                if isinstance(artifact.metadata.attributes.blobs[attr], list):
                    if index is None:
                        raise webob.exc.HTTPBadRequest(
                            explanation=_("Index is required"))
                    blob_list = getattr(artifact, attr)
                    try:
                        return blob_list[index]
                    except IndexError as e:
                        raise webob.exc.HTTPBadRequest(explanation=e.message)
                else:
                    if index is not None:
                        raise webob.exc.HTTPBadRequest(_("Not a list "
                                                         "property"))
                    return getattr(artifact, attr)
            else:
                message = _("Not a downloadable entity")
                raise webob.exc.HTTPBadRequest(explanation=message)
        except exception.Forbidden as e:
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except (glance_store.NotFound, exception.NotFound) as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        except exception.Invalid as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)

    def _restore(self, artifact_repo, artifact):
        """Restore the artifact to queued status.

        :param artifact_repo: The instance of ArtifactRepo
        :param artifact: The artifact will be restored
        """
        try:
            if artifact_repo and artifact:
                artifact.state = 'creating'
                artifact_repo.save(artifact)
        except Exception as e:
            msg = (_LE("Unable to restore artifact %(artifact_id)s: %(e)s") %
                   {'artifact_id': artifact.id,
                    'e': encodeutils.exception_to_unicode(e)})
            LOG.exception(msg)

    def list_artifact_types(self, req):
        plugins = self.plugins.plugin_map
        response = []
        base_link = "%s/v0.1/artifacts" % (CONF.public_endpoint or
                                           req.host_url)

        for type_name, plugin in six.iteritems(plugins.get("by_typename")):

            metadata = dict(
                type_name=type_name,
                displayed_name=plugin[0].metadata.type_display_name,
                versions=[]
            )

            for version in plugin:
                endpoint = version.metadata.endpoint
                type_version = "v" + version.metadata.type_version
                version_metadata = dict(
                    id=type_version,
                    link="%s/%s/%s" % (base_link, endpoint, type_version)
                )
                type_description = version.metadata.type_description
                if type_description is not None:
                    version_metadata['description'] = type_description
                metadata['versions'].append(version_metadata)
            response.append(metadata)

        return {"artifact_types": response}

    @staticmethod
    def _ensure_write_access(artifact, context):
        if context.is_admin:
            return
        if context.owner is None or context.owner != artifact.owner:
            raise exception.ArtifactForbidden(id=artifact.id)


class RequestDeserializer(wsgi.JSONRequestDeserializer,
                          jsonpatchvalidator.JsonPatchValidatorMixin):
    _available_sort_keys = ('name', 'status', 'container_format',
                            'disk_format', 'size', 'id', 'created_at',
                            'updated_at', 'version')
    _default_sort_dir = 'desc'

    _max_limit_number = 1000

    def __init__(self, schema=None, plugins=None):
        super(RequestDeserializer, self).__init__(
            methods_allowed=["replace", "remove", "add"])
        self.plugins = plugins or loader.ArtifactsPluginLoader(
            'glance.artifacts.types')

    def _validate_show_level(self, show_level):
        try:
            return Showlevel.from_str(show_level.strip().lower())
        except exception.ArtifactUnsupportedShowLevel as e:
            raise webob.exc.HTTPBadRequest(explanation=e.message)

    def show(self, req):
        res = self._process_type_from_request(req, True)
        params = req.params.copy()
        show_level = params.pop('show_level', None)
        if show_level is not None:
            res['show_level'] = self._validate_show_level(show_level)
        return res

    def _get_request_body(self, req):
        output = super(RequestDeserializer, self).default(req)
        if 'body' not in output:
            msg = _('Body expected in request.')
            raise webob.exc.HTTPBadRequest(explanation=msg)
        return output['body']

    def validate_body(self, request):
        try:
            body = self._get_request_body(request)
            return super(RequestDeserializer, self).validate_body(body)
        except exception.JsonPatchException as e:
            raise webob.exc.HTTPBadRequest(explanation=e)

    def default(self, request):
        return self._process_type_from_request(request)

    def _check_type_version(self, type_version):
        try:
            semantic_version.Version(type_version, partial=True)
        except ValueError as e:
            raise webob.exc.HTTPBadRequest(explanation=e)

    def _process_type_from_request(self, req,
                                   allow_implicit_version=False):
        try:
            type_name = req.urlvars.get('type_name')
            type_version = req.urlvars.get('type_version')
            if type_version is not None:
                self._check_type_version(type_version)
            # Even if the type_version is not specified and
            # 'allow_implicit_version' is False, this call is still needed to
            # ensure that at least one version of this type exists.
            artifact_type = self.plugins.get_class_by_endpoint(type_name,
                                                               type_version)
            res = {
                'type_name': artifact_type.metadata.type_name,
                'type_version':
                    artifact_type.metadata.type_version
                    if type_version is not None else None
            }
            if allow_implicit_version:
                res['artifact_type'] = artifact_type
            return res

        except exception.ArtifactPluginNotFound as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)

    def _validate_headers(self, req, content_type='application/json'):
        header = req.headers.get('Content-Type')
        if header != content_type:
            msg = _('Invalid headers "Content-Type": %s') % header
            raise webob.exc.HTTPBadRequest(explanation=msg)

    def create(self, req):
        self._validate_headers(req)
        res = self._process_type_from_request(req, True)
        res["artifact_data"] = self._get_request_body(req)
        return res

    def update(self, req):
        self._validate_headers(req)
        res = self._process_type_from_request(req)
        res["changes"] = self.validate_body(req)
        return res

    def update_property(self, req):
        self._validate_headers(req)
        """Data is expected in form {'data': ...}"""
        res = self._process_type_from_request(req)
        data_schema = {
            "type": "object",
            "properties": {"data": {}},
            "required": ["data"],
            "$schema": "http://json-schema.org/draft-04/schema#"}
        try:
            json_body = json.loads(req.body)
            jsonschema.validate(json_body, data_schema)
            # TODO(ivasilevskaya):
            # by now the deepest nesting level == 1 (ex. some_list/3),
            # has to be fixed for dict properties
            attr = req.urlvars["attr"]
            path_left = req.urlvars["path_left"]
            path = (attr if not path_left
                    else "%(attr)s/%(path_left)s" % {'attr': attr,
                                                     'path_left': path_left})
            res.update(data=json_body["data"], path=path)
            return res
        except (ValueError, jsonschema.ValidationError) as e:
            msg = _("Invalid json body: %s") % e.message
            raise webob.exc.HTTPBadRequest(explanation=msg)

    def upload(self, req):
        self._validate_headers(req, content_type='application/octet-stream')
        res = self._process_type_from_request(req)
        index = req.urlvars.get('path_left')
        try:
            # for blobs only one level of indexing is supported
            # (ex. bloblist/0)
            if index is not None:
                index = int(index)
        except ValueError:
            msg = _("Only list indexes are allowed for blob lists")
            raise webob.exc.HTTPBadRequest(explanation=msg)
        artifact_size = req.content_length or None
        res.update(size=artifact_size, data=req.body_file,
                   index=index)
        return res

    def download(self, req):
        res = self._process_type_from_request(req)
        index = req.urlvars.get('index')
        if index is not None:
            index = int(index)
        res.update(index=index)
        return res

    def _validate_limit(self, limit):
        if limit is None:
            return self._max_limit_number
        try:
            limit = int(limit)
        except ValueError:
            msg = _("Limit param must be an integer")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        if limit < 0:
            msg = _("Limit param must be positive")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        if limit > self._max_limit_number:
            msg = _("Limit param"
                    " must not be higher than %d") % self._max_limit_number
            raise webob.exc.HTTPBadRequest(explanation=msg)

        return limit

    def _validate_sort_key(self, sort_key, artifact_type, type_version=None):
        if sort_key in self._available_sort_keys:
            return sort_key, None
        elif type_version is None:
            msg = (_('Invalid sort key: %(sort_key)s. '
                     'If type version is not set it must be one of'
                     ' the following: %(available)s.') %
                   {'sort_key': sort_key,
                    'available': ', '.join(self._available_sort_keys)})
            raise webob.exc.HTTPBadRequest(explanation=msg)
        prop_type = artifact_type.metadata.attributes.all.get(sort_key)
        if prop_type is None or prop_type.DB_TYPE not in ['string',
                                                          'numeric',
                                                          'int',
                                                          'bool']:
            msg = (_('Invalid sort key: %(sort_key)s. '
                     'You cannot sort by this property') %
                   {'sort_key': sort_key})
            raise webob.exc.HTTPBadRequest(explanation=msg)

        return sort_key, prop_type.DB_TYPE

    def _validate_sort_dir(self, sort_dir):
        if sort_dir not in ['asc', 'desc']:
            msg = _('Invalid sort direction: %s') % sort_dir
            raise webob.exc.HTTPBadRequest(explanation=msg)

        return sort_dir

    def _get_sorting_params(self, params, artifact_type, type_version=None):

        sort_keys = []
        sort_dirs = []

        if 'sort' in params:
            for sort_param in params.pop('sort').strip().split(','):
                key, _sep, dir = sort_param.partition(':')
                if not dir:
                    dir = self._default_sort_dir
                sort_keys.append(self._validate_sort_key(key.strip(),
                                                         artifact_type,
                                                         type_version))
                sort_dirs.append(self._validate_sort_dir(dir.strip()))

        if not sort_keys:
            sort_keys = [('created_at', None)]
        if not sort_dirs:
            sort_dirs = [self._default_sort_dir]

        return sort_keys, sort_dirs

    def _bring_to_type(self, type_name, value):
        mapper = {'int': int,
                  'string': str,
                  'text': str,
                  'bool': bool,
                  'numeric': float}
        return mapper[type_name](value)

    def _get_filters(self, artifact_type, params):
        error_msg = 'Unexpected filter property'
        filters = dict()
        for filter, raw_value in params.items():

            # first, get the comparison operator
            left, sep, right = raw_value.strip().partition(':')
            if not sep:
                op = "default"
                value = left.strip()
            else:
                op = left.strip().upper()
                value = right.strip()

            # then, understand what's the property to filter and its value
            if '.' in filter:  # Indicates a dict-valued property with a key
                prop_name, key = filter.split('.', 1)
            else:
                prop_name = filter
                key = None
            prop_type = artifact_type.metadata.attributes.all.get(prop_name)
            if prop_type is None:
                raise webob.exc.HTTPBadRequest(error_msg)
            key_only_check = False
            position = None
            if isinstance(prop_type, dict):
                if key is None:
                    key = value
                    val = None
                    key_only_check = True
                else:
                    val = value

                if isinstance(prop_type.properties, dict):
                    # This one is to handle the case of composite dict, having
                    # different types of values at different keys, i.e. object
                    prop_type = prop_type.properties.get(key)
                    if prop_type is None:
                        raise webob.exc.HTTPBadRequest(error_msg)
                else:
                    prop_type = prop_type.properties

                property_name = prop_name + '.' + key
                property_value = val
            else:
                if key is not None:
                    raise webob.exc.HTTPBadRequest(error_msg)
                property_name = prop_name
                property_value = value

            # now detect the value DB type
            if prop_type.DB_TYPE is not None:
                str_type = prop_type.DB_TYPE
            elif isinstance(prop_type, list):
                if not isinstance(prop_type.item_type, list):
                    position = "any"
                    str_type = prop_type.item_type.DB_TYPE
                else:
                    raise webob.exc.HTTPBadRequest('Filtering by tuple-like'
                                                   ' fields is not supported')
            else:
                raise webob.exc.HTTPBadRequest(error_msg)
            if property_value is not None:
                property_value = self._bring_to_type(str_type, property_value)

            # convert the default operation to NE, EQ or IN
            if key_only_check:
                if op == 'default':
                    op = 'NE'
                else:
                    raise webob.exc.HTTPBadRequest('Comparison not supported '
                                                   'for key-only filtering')
            else:
                if op == 'default':
                    op = 'IN' if isinstance(prop_type, list) else 'EQ'

            filters.setdefault(property_name, [])

            filters[property_name].append(dict(operator=op, position=position,
                                               value=property_value,
                                               type=str_type))
        return filters

    def list(self, req):
        res = self._process_type_from_request(req, True)
        params = req.params.copy()
        show_level = params.pop('show_level', None)
        if show_level is not None:
            res['show_level'] = self._validate_show_level(show_level.strip())

        limit = params.pop('limit', None)
        marker = params.pop('marker', None)

        query_params = dict()

        query_params['sort_keys'], query_params['sort_dirs'] = (
            self._get_sorting_params(params, res['artifact_type'],
                                     res['type_version']))

        if marker is not None:
            query_params['marker'] = marker

        query_params['limit'] = self._validate_limit(limit)

        query_params['filters'] = self._get_filters(res['artifact_type'],
                                                    params)

        query_params['type_name'] = res['artifact_type'].metadata.type_name

        return query_params

    def list_artifact_types(self, req):
        return {}


class ResponseSerializer(wsgi.JSONResponseSerializer):
    # TODO(ivasilevskaya): ideally this should be autogenerated/loaded
    ARTIFACTS_ENDPOINT = '/v0.1/artifacts'
    fields = ['id', 'name', 'version', 'type_name', 'type_version',
              'visibility', 'state', 'owner', 'scope', 'created_at',
              'updated_at', 'tags', 'dependencies', 'blobs', 'properties']

    def __init__(self, schema=None):
        super(ResponseSerializer, self).__init__()

    def default(self, response, res):
        artifact = serialization.serialize_for_client(
            res, show_level=Showlevel.DIRECT)
        body = json.dumps(artifact, ensure_ascii=False)
        response.unicode_body = six.text_type(body)
        response.content_type = 'application/json'

    def create(self, response, artifact):
        response.status_int = 201
        self.default(response, artifact)
        response.location = (
            '%(root_url)s/%(type_name)s/v%(type_version)s/%(id)s' % dict(
                root_url=ResponseSerializer.ARTIFACTS_ENDPOINT,
                type_name=artifact.metadata.endpoint,
                type_version=artifact.metadata.type_version,
                id=artifact.id))

    def list(self, response, res):
        params = dict(response.request.params)
        params.pop('marker', None)
        query = urlparse.urlencode(params)
        type_name = response.request.urlvars.get('type_name')
        type_version = response.request.urlvars.get('type_version')
        if response.request.urlvars.get('state') == 'creating':
            drafts = "/drafts"
        else:
            drafts = ""

        artifacts_list = [
            serialization.serialize_for_client(a, show_level=Showlevel.NONE)
            for a in res['artifacts']]
        url = "/v0.1/artifacts"
        if type_name:
            url += "/" + type_name
        if type_version:
            url += "/v" + type_version
        url += drafts
        if query:
            first_url = url + "?" + query
        else:
            first_url = url
        body = {
            "artifacts": artifacts_list,
            "first": first_url
        }
        if 'next_marker' in res:
            params['marker'] = res['next_marker']
            next_query = urlparse.urlencode(params)
            body['next'] = url + '?' + next_query
        content = json.dumps(body, ensure_ascii=False)
        response.unicode_body = six.text_type(content)
        response.content_type = 'application/json'

    def delete(self, response, result):
        response.status_int = 204

    def download(self, response, blob):
        response.headers['Content-Type'] = 'application/octet-stream'
        response.app_iter = iter(blob.data_stream)
        if blob.checksum:
            response.headers['Content-MD5'] = blob.checksum
        response.headers['Content-Length'] = str(blob.size)

    def list_artifact_types(self, response, res):
        body = json.dumps(res, ensure_ascii=False)
        response.unicode_body = six.text_type(body)
        response.content_type = 'application/json'


def create_resource():
    """Images resource factory method"""
    plugins = loader.ArtifactsPluginLoader('glance.artifacts.types')
    deserializer = RequestDeserializer(plugins=plugins)
    serializer = ResponseSerializer()
    controller = ArtifactsController(plugins=plugins)
    return wsgi.Resource(controller, deserializer, serializer)



# Copyright 2011 OpenStack Foundation
# Copyright 2015 Mirantis, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import tempfile

import six
import webob

from glance.common import exception
from glance.common import utils
from glance.tests import utils as test_utils


class TestUtils(test_utils.BaseTestCase):
    """Test routines in glance.utils"""

    def test_cooperative_reader(self):
        """Ensure cooperative reader class accesses all bytes of file"""
        BYTES = 1024
        bytes_read = 0
        with tempfile.TemporaryFile('w+') as tmp_fd:
            tmp_fd.write('*' * BYTES)
            tmp_fd.seek(0)
            for chunk in utils.CooperativeReader(tmp_fd):
                bytes_read += len(chunk)

        self.assertEqual(BYTES, bytes_read)

        bytes_read = 0
        with tempfile.TemporaryFile('w+') as tmp_fd:
            tmp_fd.write('*' * BYTES)
            tmp_fd.seek(0)
            reader = utils.CooperativeReader(tmp_fd)
            byte = reader.read(1)
            while len(byte) != 0:
                bytes_read += 1
                byte = reader.read(1)

        self.assertEqual(BYTES, bytes_read)

    def test_cooperative_reader_of_iterator(self):
        """Ensure cooperative reader supports iterator backends too"""
        data = b'abcdefgh'
        data_list = [data[i:i + 1] * 3 for i in range(len(data))]
        reader = utils.CooperativeReader(data_list)
        chunks = []
        while True:
            chunks.append(reader.read(3))
            if chunks[-1] == b'':
                break
        meat = b''.join(chunks)
        self.assertEqual(b'aaabbbcccdddeeefffggghhh', meat)

    def test_cooperative_reader_of_iterator_stop_iteration_err(self):
        """Ensure cooperative reader supports iterator backends too"""
        reader = utils.CooperativeReader([l * 3 for l in ''])
        chunks = []
        while True:
            chunks.append(reader.read(3))
            if chunks[-1] == b'':
                break
        meat = b''.join(chunks)
        self.assertEqual(b'', meat)

    def _create_generator(self, chunk_size, max_iterations):
        chars = b'abc'
        iteration = 0
        while True:
            index = iteration % len(chars)
            chunk = chars[index:index + 1] * chunk_size
            yield chunk
            iteration += 1
            if iteration >= max_iterations:
                raise StopIteration()

    def _test_reader_chunked(self, chunk_size, read_size, max_iterations=5):
        generator = self._create_generator(chunk_size, max_iterations)
        reader = utils.CooperativeReader(generator)
        result = bytearray()
        while True:
            data = reader.read(read_size)
            if len(data) == 0:
                break
            self.assertLessEqual(len(data), read_size)
            result += data
        expected = (b'a' * chunk_size +
                    b'b' * chunk_size +
                    b'c' * chunk_size +
                    b'a' * chunk_size +
                    b'b' * chunk_size)
        self.assertEqual(expected, bytes(result))

    def test_cooperative_reader_preserves_size_chunk_less_then_read(self):
        self._test_reader_chunked(43, 101)

    def test_cooperative_reader_preserves_size_chunk_equals_read(self):
        self._test_reader_chunked(1024, 1024)

    def test_cooperative_reader_preserves_size_chunk_more_then_read(self):
        chunk_size = 16 * 1024 * 1024  # 16 Mb, as in remote http source
        read_size = 8 * 1024           # 8k, as in httplib
        self._test_reader_chunked(chunk_size, read_size)

    def test_limiting_reader(self):
        """Ensure limiting reader class accesses all bytes of file"""
        BYTES = 1024
        bytes_read = 0
        data = six.StringIO("*" * BYTES)
        for chunk in utils.LimitingReader(data, BYTES):
            bytes_read += len(chunk)

        self.assertEqual(BYTES, bytes_read)

        bytes_read = 0
        data = six.StringIO("*" * BYTES)
        reader = utils.LimitingReader(data, BYTES)
        byte = reader.read(1)
        while len(byte) != 0:
            bytes_read += 1
            byte = reader.read(1)

        self.assertEqual(BYTES, bytes_read)

    def test_limiting_reader_fails(self):
        """Ensure limiting reader class throws exceptions if limit exceeded"""
        BYTES = 1024

        def _consume_all_iter():
            bytes_read = 0
            data = six.StringIO("*" * BYTES)
            for chunk in utils.LimitingReader(data, BYTES - 1):
                bytes_read += len(chunk)

        self.assertRaises(exception.ImageSizeLimitExceeded, _consume_all_iter)

        def _consume_all_read():
            bytes_read = 0
            data = six.StringIO("*" * BYTES)
            reader = utils.LimitingReader(data, BYTES - 1)
            byte = reader.read(1)
            while len(byte) != 0:
                bytes_read += 1
                byte = reader.read(1)

        self.assertRaises(exception.ImageSizeLimitExceeded, _consume_all_read)

    def test_get_meta_from_headers(self):
        resp = webob.Response()
        resp.headers = {"x-image-meta-name": 'test',
                        'x-image-meta-virtual-size': 80}
        result = utils.get_image_meta_from_headers(resp)
        self.assertEqual({'name': 'test', 'properties': {},
                          'virtual_size': 80}, result)

    def test_get_meta_from_headers_none_virtual_size(self):
        resp = webob.Response()
        resp.headers = {"x-image-meta-name": 'test',
                        'x-image-meta-virtual-size': 'None'}
        result = utils.get_image_meta_from_headers(resp)
        self.assertEqual({'name': 'test', 'properties': {},
                          'virtual_size': None}, result)

    def test_get_meta_from_headers_bad_headers(self):
        resp = webob.Response()
        resp.headers = {"x-image-meta-bad": 'test'}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          utils.get_image_meta_from_headers, resp)
        resp.headers = {"x-image-meta-": 'test'}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          utils.get_image_meta_from_headers, resp)
        resp.headers = {"x-image-meta-*": 'test'}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          utils.get_image_meta_from_headers, resp)

    def test_image_meta(self):
        image_meta = {'x-image-meta-size': 'test'}
        image_meta_properties = {'properties': {'test': "test"}}
        actual = utils.image_meta_to_http_headers(image_meta)
        actual_test2 = utils.image_meta_to_http_headers(
            image_meta_properties)
        self.assertEqual({'x-image-meta-x-image-meta-size': u'test'}, actual)
        self.assertEqual({'x-image-meta-property-test': u'test'},
                         actual_test2)

    def test_create_mashup_dict_with_different_core_custom_properties(self):
        image_meta = {
            'id': 'test-123',
            'name': 'fake_image',
            'status': 'active',
            'created_at': '',
            'min_disk': '10G',
            'min_ram': '1024M',
            'protected': False,
            'locations': '',
            'checksum': 'c1234',
            'owner': '',
            'disk_format': 'raw',
            'container_format': 'bare',
            'size': '123456789',
            'virtual_size': '123456789',
            'is_public': 'public',
            'deleted': True,
            'updated_at': '',
            'properties': {'test_key': 'test_1234'},
        }

        mashup_dict = utils.create_mashup_dict(image_meta)
        self.assertNotIn('properties', mashup_dict)
        self.assertEqual(image_meta['properties']['test_key'],
                         mashup_dict['test_key'])

    def test_create_mashup_dict_with_same_core_custom_properties(self):
        image_meta = {
            'id': 'test-123',
            'name': 'fake_image',
            'status': 'active',
            'created_at': '',
            'min_disk': '10G',
            'min_ram': '1024M',
            'protected': False,
            'locations': '',
            'checksum': 'c1234',
            'owner': '',
            'disk_format': 'raw',
            'container_format': 'bare',
            'size': '123456789',
            'virtual_size': '123456789',
            'is_public': 'public',
            'deleted': True,
            'updated_at': '',
            'properties': {'min_ram': '2048M'},
        }

        mashup_dict = utils.create_mashup_dict(image_meta)
        self.assertNotIn('properties', mashup_dict)
        self.assertNotEqual(image_meta['properties']['min_ram'],
                            mashup_dict['min_ram'])
        self.assertEqual(image_meta['min_ram'], mashup_dict['min_ram'])

    def test_mutating(self):
        class FakeContext(object):
            def __init__(self):
                self.read_only = False

        class Fake(object):
            def __init__(self):
                self.context = FakeContext()

        def fake_function(req, context):
            return 'test passed'

        req = webob.Request.blank('/some_request')
        result = utils.mutating(fake_function)
        self.assertEqual("test passed", result(req, Fake()))

    def test_validate_key_cert_key(self):
        self.config(digest_algorithm='sha256')
        var_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                               '../../', 'var'))
        keyfile = os.path.join(var_dir, 'privatekey.key')
        certfile = os.path.join(var_dir, 'certificate.crt')
        utils.validate_key_cert(keyfile, certfile)

    def test_validate_key_cert_no_private_key(self):
        with tempfile.NamedTemporaryFile('w+') as tmpf:
            self.assertRaises(RuntimeError,
                              utils.validate_key_cert,
                              "/not/a/file", tmpf.name)

    def test_validate_key_cert_cert_cant_read(self):
        with tempfile.NamedTemporaryFile('w+') as keyf:
            with tempfile.NamedTemporaryFile('w+') as certf:
                os.chmod(certf.name, 0)
                self.assertRaises(RuntimeError,
                                  utils.validate_key_cert,
                                  keyf.name, certf.name)

    def test_validate_key_cert_key_cant_read(self):
        with tempfile.NamedTemporaryFile('w+') as keyf:
            with tempfile.NamedTemporaryFile('w+') as certf:
                os.chmod(keyf.name, 0)
                self.assertRaises(RuntimeError,
                                  utils.validate_key_cert,
                                  keyf.name, certf.name)

    def test_invalid_digest_algorithm(self):
        self.config(digest_algorithm='fake_algorithm')
        var_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                               '../../', 'var'))
        keyfile = os.path.join(var_dir, 'privatekey.key')
        certfile = os.path.join(var_dir, 'certificate.crt')
        self.assertRaises(ValueError,
                          utils.validate_key_cert,
                          keyfile, certfile)

    def test_valid_hostname(self):
        valid_inputs = ['localhost',
                        'glance04-a'
                        'G',
                        '528491']

        for input_str in valid_inputs:
            self.assertTrue(utils.is_valid_hostname(input_str))

    def test_valid_hostname_fail(self):
        invalid_inputs = ['localhost.localdomain',
                          '192.168.0.1',
                          u'\u2603',
                          'glance02.stack42.local']

        for input_str in invalid_inputs:
            self.assertFalse(utils.is_valid_hostname(input_str))

    def test_valid_fqdn(self):
        valid_inputs = ['localhost.localdomain',
                        'glance02.stack42.local'
                        'glance04-a.stack47.local',
                        'img83.glance.xn--penstack-r74e.org']

        for input_str in valid_inputs:
            self.assertTrue(utils.is_valid_fqdn(input_str))

    def test_valid_fqdn_fail(self):
        invalid_inputs = ['localhost',
                          '192.168.0.1',
                          '999.88.77.6',
                          u'\u2603.local',
                          'glance02.stack42']

        for input_str in invalid_inputs:
            self.assertFalse(utils.is_valid_fqdn(input_str))

    def test_valid_host_port_string(self):
        valid_pairs = ['10.11.12.13:80',
                       '172.17.17.1:65535',
                       '[fe80::a:b:c:d]:9990',
                       'localhost:9990',
                       'localhost.localdomain:9990',
                       'glance02.stack42.local:1234',
                       'glance04-a.stack47.local:1234',
                       'img83.glance.xn--penstack-r74e.org:13080']

        for pair_str in valid_pairs:
            host, port = utils.parse_valid_host_port(pair_str)

            escaped = pair_str.startswith('[')
            expected_host = '%s%s%s' % ('[' if escaped else '', host,
                                        ']' if escaped else '')

            self.assertTrue(pair_str.startswith(expected_host))
            self.assertGreater(port, 0)

            expected_pair = '%s:%d' % (expected_host, port)
            self.assertEqual(expected_pair, pair_str)

    def test_valid_host_port_string_fail(self):
        invalid_pairs = ['',
                         '10.11.12.13',
                         '172.17.17.1:99999',
                         '290.12.52.80:5673',
                         'absurd inputs happen',
                         u'\u2601',
                         u'\u2603:8080',
                         'fe80::1',
                         '[fe80::2]',
                         '<fe80::3>:5673',
                         '[fe80::a:b:c:d]9990',
                         'fe80:a:b:c:d:e:f:1:2:3:4',
                         'fe80:a:b:c:d:e:f:g',
                         'fe80::1:8080',
                         '[fe80:a:b:c:d:e:f:g]:9090',
                         '[a:b:s:u:r:d]:fe80']

        for pair in invalid_pairs:
            self.assertRaises(ValueError,
                              utils.parse_valid_host_port,
                              pair)


class SplitFilterOpTestCase(test_utils.BaseTestCase):

    def test_less_than_operator(self):
        expr = 'lt:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('lt', 'bar'), returned)

    def test_less_than_equal_operator(self):
        expr = 'lte:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('lte', 'bar'), returned)

    def test_greater_than_operator(self):
        expr = 'gt:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('gt', 'bar'), returned)

    def test_greater_than_equal_operator(self):
        expr = 'gte:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('gte', 'bar'), returned)

    def test_not_equal_operator(self):
        expr = 'neq:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('neq', 'bar'), returned)

    def test_equal_operator(self):
        expr = 'eq:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('eq', 'bar'), returned)

    def test_in_operator(self):
        expr = 'in:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('in', 'bar'), returned)

    def test_split_filter_value_for_quotes(self):
        expr = '\"fake\\\"name\",fakename,\"fake,name\"'
        returned = utils.split_filter_value_for_quotes(expr)
        list_values = ['fake\\"name', 'fakename', 'fake,name']
        self.assertEqual(list_values, returned)

    def test_validate_quotes(self):
        expr = '\"aaa\\\"aa\",bb,\"cc\"'
        returned = utils.validate_quotes(expr)
        self.assertIsNone(returned)

        invalid_expr = ['\"aa', 'ss\"', 'aa\"bb\"cc', '\"aa\"\"bb\"']
        for expr in invalid_expr:
            self.assertRaises(exception.InvalidParameterValue,
                              utils.validate_quotes,
                              expr)

    def test_default_operator(self):
        expr = 'bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('eq', expr), returned)

    def test_default_operator_with_datetime(self):
        expr = '2015-08-27T09:49:58Z'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('eq', expr), returned)

    def test_operator_with_datetime(self):
        expr = 'lt:2015-08-27T09:49:58Z'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('lt', '2015-08-27T09:49:58Z'), returned)


class EvaluateFilterOpTestCase(test_utils.BaseTestCase):

    def test_less_than_operator(self):
        self.assertTrue(utils.evaluate_filter_op(9, 'lt', 10))
        self.assertFalse(utils.evaluate_filter_op(10, 'lt', 10))
        self.assertFalse(utils.evaluate_filter_op(11, 'lt', 10))

    def test_less_than_equal_operator(self):
        self.assertTrue(utils.evaluate_filter_op(9, 'lte', 10))
        self.assertTrue(utils.evaluate_filter_op(10, 'lte', 10))
        self.assertFalse(utils.evaluate_filter_op(11, 'lte', 10))

    def test_greater_than_operator(self):
        self.assertFalse(utils.evaluate_filter_op(9, 'gt', 10))
        self.assertFalse(utils.evaluate_filter_op(10, 'gt', 10))
        self.assertTrue(utils.evaluate_filter_op(11, 'gt', 10))

    def test_greater_than_equal_operator(self):
        self.assertFalse(utils.evaluate_filter_op(9, 'gte', 10))
        self.assertTrue(utils.evaluate_filter_op(10, 'gte', 10))
        self.assertTrue(utils.evaluate_filter_op(11, 'gte', 10))

    def test_not_equal_operator(self):
        self.assertTrue(utils.evaluate_filter_op(9, 'neq', 10))
        self.assertFalse(utils.evaluate_filter_op(10, 'neq', 10))
        self.assertTrue(utils.evaluate_filter_op(11, 'neq', 10))

    def test_equal_operator(self):
        self.assertFalse(utils.evaluate_filter_op(9, 'eq', 10))
        self.assertTrue(utils.evaluate_filter_op(10, 'eq', 10))
        self.assertFalse(utils.evaluate_filter_op(11, 'eq', 10))

    def test_invalid_operator(self):
        self.assertRaises(exception.InvalidFilterOperatorValue,
                          utils.evaluate_filter_op, '10', 'bar', '8')



# Copyright 2011-2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from glance.common.glare import definitions


class BaseArtifact(definitions.ArtifactType):
    __type_version__ = "1.0"
    prop1 = definitions.String()
    prop2 = definitions.Integer()
    int_list = definitions.Array(item_type=definitions.Integer(max_value=10,
                                                               min_value=1))
    depends_on = definitions.ArtifactReference(type_name='MyArtifact')
    references = definitions.ArtifactReferenceList()

    image_file = definitions.BinaryObject()
    screenshots = definitions.BinaryObjectList()



# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
import os
import uuid

from mock import patch
from six.moves import reload_module
import testtools

from glance.api.v1.images import Controller as acontroller
from glance.common import client as test_client
from glance.common import config
from glance.common import exception
from glance.common import timeutils
from glance import context
from glance.db.sqlalchemy import api as db_api
from glance.registry.api.v1.images import Controller as rcontroller
import glance.registry.client.v1.api as rapi
from glance.registry.client.v1.api import client as rclient
from glance.tests.unit import base
from glance.tests import utils as test_utils
import webob

_gen_uuid = lambda: str(uuid.uuid4())

UUID1 = _gen_uuid()
UUID2 = _gen_uuid()

# NOTE(bcwaldon): needed to init config_dir cli opt
config.parse_args(args=[])


class TestRegistryV1Client(base.IsolatedUnitTest, test_utils.RegistryAPIMixIn):

    """
    Test proper actions made for both valid and invalid requests
    against a Registry service
    """

    def setUp(self):
        """Establish a clean test environment"""
        super(TestRegistryV1Client, self).setUp()
        db_api.get_engine()
        self.context = context.RequestContext(is_admin=True)

        self.FIXTURES = [
            self.get_fixture(
                id=UUID1, name='fake image #1', is_public=False,
                disk_format='ami', container_format='ami', size=13,
                location="swift://user:passwd@acct/container/obj.tar.0",
                properties={'type': 'kernel'}),
            self.get_fixture(id=UUID2, name='fake image #2', properties={},
                             size=19, location="file:///tmp/glance-tests/2")]
        self.destroy_fixtures()
        self.create_fixtures()
        self.client = rclient.RegistryClient("0.0.0.0")

    def tearDown(self):
        """Clear the test environment"""
        super(TestRegistryV1Client, self).tearDown()
        self.destroy_fixtures()

    def test_get_image_index(self):
        """Test correct set of public image returned"""
        fixture = {
            'id': UUID2,
            'name': 'fake image #2'
        }
        images = self.client.get_images()
        self.assertEqualImages(images, (UUID2,), unjsonify=False)

        for k, v in fixture.items():
            self.assertEqual(v, images[0][k])

    def test_create_image_with_null_min_disk_min_ram(self):
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, min_disk=None, min_ram=None)
        db_api.image_create(self.context, extra_fixture)
        image = self.client.get_image(UUID3)
        self.assertEqual(0, image["min_ram"])
        self.assertEqual(0, image["min_disk"])

    def test_get_index_sort_name_asc(self):
        """
        Tests that the /images registry API returns list of
        public images sorted alphabetically by name in
        ascending order.
        """
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='asdf')

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='xyz')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(sort_key='name', sort_dir='asc')

        self.assertEqualImages(images, (UUID3, UUID2, UUID4), unjsonify=False)

    def test_get_index_sort_status_desc(self):
        """
        Tests that the /images registry API returns list of
        public images sorted alphabetically by status in
        descending order.
        """
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='asdf',
                                         status='queued')

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='xyz')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(sort_key='status', sort_dir='desc')

        self.assertEqualImages(images, (UUID3, UUID4, UUID2), unjsonify=False)

    def test_get_index_sort_disk_format_asc(self):
        """
        Tests that the /images registry API returns list of
        public images sorted alphabetically by disk_format in
        ascending order.
        """
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='asdf',
                                         disk_format='ami',
                                         container_format='ami')

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='xyz',
                                         disk_format='vdi')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(sort_key='disk_format',
                                        sort_dir='asc')

        self.assertEqualImages(images, (UUID3, UUID4, UUID2), unjsonify=False)

    def test_get_index_sort_container_format_desc(self):
        """
        Tests that the /images registry API returns list of
        public images sorted alphabetically by container_format in
        descending order.
        """
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='asdf',
                                         disk_format='ami',
                                         container_format='ami')

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='xyz',
                                         disk_format='iso',
                                         container_format='bare')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(sort_key='container_format',
                                        sort_dir='desc')

        self.assertEqualImages(images, (UUID2, UUID4, UUID3), unjsonify=False)

    def test_get_index_sort_size_asc(self):
        """
        Tests that the /images registry API returns list of
        public images sorted by size in ascending order.
        """
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='asdf',
                                         disk_format='ami',
                                         container_format='ami', size=100)

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='asdf',
                                         disk_format='iso',
                                         container_format='bare', size=2)

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(sort_key='size', sort_dir='asc')

        self.assertEqualImages(images, (UUID4, UUID2, UUID3), unjsonify=False)

    def test_get_index_sort_created_at_asc(self):
        """
        Tests that the /images registry API returns list of
        public images sorted by created_at in ascending order.
        """
        now = timeutils.utcnow()
        time1 = now + datetime.timedelta(seconds=5)
        time2 = now

        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, created_at=time1)

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, created_at=time2)

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(sort_key='created_at', sort_dir='asc')

        self.assertEqualImages(images, (UUID2, UUID4, UUID3), unjsonify=False)

    def test_get_index_sort_updated_at_desc(self):
        """
        Tests that the /images registry API returns list of
        public images sorted by updated_at in descending order.
        """
        now = timeutils.utcnow()
        time1 = now + datetime.timedelta(seconds=5)
        time2 = now

        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, created_at=None,
                                         updated_at=time1)

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, created_at=None,
                                         updated_at=time2)

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(sort_key='updated_at', sort_dir='desc')

        self.assertEqualImages(images, (UUID3, UUID4, UUID2), unjsonify=False)

    def test_get_image_index_marker(self):
        """Test correct set of images returned with marker param."""
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='new name! #123',
                                         status='saving')

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='new name! #125',
                                         status='saving')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(marker=UUID4)

        self.assertEqualImages(images, (UUID3, UUID2), unjsonify=False)

    def test_get_image_index_invalid_marker(self):
        """Test exception is raised when marker is invalid"""
        self.assertRaises(exception.Invalid,
                          self.client.get_images,
                          marker=_gen_uuid())

    def test_get_image_index_forbidden_marker(self):
        """Test exception is raised when marker is forbidden"""
        UUID5 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID5, owner='0123',
                                         status='saving', is_public=False)

        db_api.image_create(self.context, extra_fixture)

        def non_admin_get_images(self, context, *args, **kwargs):
            """Convert to non-admin context"""
            context.is_admin = False
            rcontroller.__get_images(self, context, *args, **kwargs)

        rcontroller.__get_images = rcontroller._get_images
        self.stubs.Set(rcontroller, '_get_images', non_admin_get_images)
        self.assertRaises(exception.Invalid,
                          self.client.get_images,
                          marker=UUID5)

    def test_get_image_index_private_marker(self):
        """Test exception is not raised if private non-owned marker is used"""
        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, owner='1234',
                                         status='saving', is_public=False)

        db_api.image_create(self.context, extra_fixture)

        try:
            self.client.get_images(marker=UUID4)
        except Exception as e:
            self.fail("Unexpected exception '%s'" % e)

    def test_get_image_index_limit(self):
        """Test correct number of images returned with limit param."""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(limit=2)
        self.assertEqual(2, len(images))

    def test_get_image_index_marker_limit(self):
        """Test correct set of images returned with marker/limit params."""
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='new name! #123',
                                         status='saving')

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='new name! #125',
                                         status='saving')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(marker=UUID3, limit=1)

        self.assertEqualImages(images, (UUID2,), unjsonify=False)

    def test_get_image_index_limit_None(self):
        """Test correct set of images returned with limit param == None."""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(limit=None)
        self.assertEqual(3, len(images))

    def test_get_image_index_by_name(self):
        """
        Test correct set of public, name-filtered image returned. This
        is just a sanity check, we test the details call more in-depth.
        """
        extra_fixture = self.get_fixture(id=_gen_uuid(), name='new name! #123')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images(filters={'name': 'new name! #123'})
        self.assertEqual(1, len(images))

        for image in images:
            self.assertEqual('new name! #123', image['name'])

    def test_get_image_details(self):
        """Tests that the detailed info about public images returned"""
        fixture = self.get_fixture(id=UUID2, name='fake image #2',
                                   properties={}, size=19)

        images = self.client.get_images_detailed()

        self.assertEqual(1, len(images))
        for k, v in fixture.items():
            self.assertEqual(v, images[0][k])

    def test_get_image_details_marker_limit(self):
        """Test correct set of images returned with marker/limit params."""
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, status='saving')

        db_api.image_create(self.context, extra_fixture)

        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images_detailed(marker=UUID3, limit=1)

        self.assertEqualImages(images, (UUID2,), unjsonify=False)

    def test_get_image_details_invalid_marker(self):
        """Test exception is raised when marker is invalid"""
        self.assertRaises(exception.Invalid,
                          self.client.get_images_detailed,
                          marker=_gen_uuid())

    def test_get_image_details_forbidden_marker(self):
        """Test exception is raised when marker is forbidden"""
        UUID5 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID5, is_public=False,
                                         owner='0123', status='saving')

        db_api.image_create(self.context, extra_fixture)

        def non_admin_get_images(self, context, *args, **kwargs):
            """Convert to non-admin context"""
            context.is_admin = False
            rcontroller.__get_images(self, context, *args, **kwargs)

        rcontroller.__get_images = rcontroller._get_images
        self.stubs.Set(rcontroller, '_get_images', non_admin_get_images)
        self.assertRaises(exception.Invalid,
                          self.client.get_images_detailed,
                          marker=UUID5)

    def test_get_image_details_private_marker(self):
        """Test exception is not raised if private non-owned marker is used"""
        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, is_public=False,
                                         owner='1234', status='saving')

        db_api.image_create(self.context, extra_fixture)

        try:
            self.client.get_images_detailed(marker=UUID4)
        except Exception as e:
            self.fail("Unexpected exception '%s'" % e)

    def test_get_image_details_by_name(self):
        """Tests that a detailed call can be filtered by name"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), name='new name! #123')

        db_api.image_create(self.context, extra_fixture)

        filters = {'name': 'new name! #123'}
        images = self.client.get_images_detailed(filters=filters)

        self.assertEqual(1, len(images))
        for image in images:
            self.assertEqual('new name! #123', image['name'])

    def test_get_image_details_by_status(self):
        """Tests that a detailed call can be filtered by status"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images_detailed(filters={'status': 'saving'})

        self.assertEqual(1, len(images))
        for image in images:
            self.assertEqual('saving', image['status'])

    def test_get_image_details_by_container_format(self):
        """Tests that a detailed call can be filtered by container_format"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        filters = {'container_format': 'ovf'}
        images = self.client.get_images_detailed(filters=filters)

        self.assertEqual(2, len(images))
        for image in images:
            self.assertEqual('ovf', image['container_format'])

    def test_get_image_details_by_disk_format(self):
        """Tests that a detailed call can be filtered by disk_format"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        filters = {'disk_format': 'vhd'}
        images = self.client.get_images_detailed(filters=filters)

        self.assertEqual(2, len(images))
        for image in images:
            self.assertEqual('vhd', image['disk_format'])

    def test_get_image_details_with_maximum_size(self):
        """Tests that a detailed call can be filtered by size_max"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving',
                                         size=21)

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images_detailed(filters={'size_max': 20})

        self.assertEqual(1, len(images))
        for image in images:
            self.assertLessEqual(image['size'], 20)

    def test_get_image_details_with_minimum_size(self):
        """Tests that a detailed call can be filtered by size_min"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images_detailed(filters={'size_min': 20})

        self.assertEqual(1, len(images))
        for image in images:
            self.assertGreaterEqual(image['size'], 20)

    def test_get_image_details_with_changes_since(self):
        """Tests that a detailed call can be filtered by changes-since"""
        dt1 = timeutils.utcnow() - datetime.timedelta(1)
        iso1 = timeutils.isotime(dt1)

        dt2 = timeutils.utcnow() + datetime.timedelta(1)
        iso2 = timeutils.isotime(dt2)

        dt3 = timeutils.utcnow() + datetime.timedelta(2)

        dt4 = timeutils.utcnow() + datetime.timedelta(3)
        iso4 = timeutils.isotime(dt4)

        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='fake image #3')

        db_api.image_create(self.context, extra_fixture)
        db_api.image_destroy(self.context, UUID3)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='fake image #4',
                                         created_at=dt3, updated_at=dt3)

        db_api.image_create(self.context, extra_fixture)

        # Check a standard list, 4 images in db (2 deleted)
        images = self.client.get_images_detailed(filters={})
        self.assertEqualImages(images, (UUID4, UUID2), unjsonify=False)

        # Expect 3 images (1 deleted)
        filters = {'changes-since': iso1}
        images = self.client.get_images(filters=filters)
        self.assertEqualImages(images, (UUID4, UUID3, UUID2), unjsonify=False)

        # Expect 1 images (0 deleted)
        filters = {'changes-since': iso2}
        images = self.client.get_images_detailed(filters=filters)
        self.assertEqualImages(images, (UUID4,), unjsonify=False)

        # Expect 0 images (0 deleted)
        filters = {'changes-since': iso4}
        images = self.client.get_images(filters=filters)
        self.assertEqualImages(images, (), unjsonify=False)

    def test_get_image_details_with_size_min(self):
        """Tests that a detailed call can be filtered by size_min"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images_detailed(filters={'size_min': 20})
        self.assertEqual(1, len(images))

        for image in images:
            self.assertGreaterEqual(image['size'], 20)

    def test_get_image_details_by_property(self):
        """Tests that a detailed call can be filtered by a property"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving',
                                         properties={'p a': 'v a'})

        db_api.image_create(self.context, extra_fixture)

        filters = {'property-p a': 'v a'}
        images = self.client.get_images_detailed(filters=filters)
        self.assertEqual(1, len(images))

        for image in images:
            self.assertEqual('v a', image['properties']['p a'])

    def test_get_image_is_public_v1(self):
        """Tests that a detailed call can be filtered by a property"""
        extra_fixture = self.get_fixture(id=_gen_uuid(), status='saving',
                                         properties={'is_public': 'avalue'})

        context = copy.copy(self.context)
        db_api.image_create(context, extra_fixture)

        filters = {'property-is_public': 'avalue'}
        images = self.client.get_images_detailed(filters=filters)
        self.assertEqual(1, len(images))

        for image in images:
            self.assertEqual('avalue', image['properties']['is_public'])

    def test_get_image_details_sort_disk_format_asc(self):
        """
        Tests that a detailed call returns list of
        public images sorted alphabetically by disk_format in
        ascending order.
        """
        UUID3 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID3, name='asdf',
                                         disk_format='ami',
                                         container_format='ami')

        db_api.image_create(self.context, extra_fixture)

        UUID4 = _gen_uuid()
        extra_fixture = self.get_fixture(id=UUID4, name='xyz',
                                         disk_format='vdi')

        db_api.image_create(self.context, extra_fixture)

        images = self.client.get_images_detailed(sort_key='disk_format',
                                                 sort_dir='asc')

        self.assertEqualImages(images, (UUID3, UUID4, UUID2), unjsonify=False)

    def test_get_image(self):
        """Tests that the detailed info about an image returned"""
        fixture = self.get_fixture(id=UUID1, name='fake image #1',
                                   disk_format='ami', container_format='ami',
                                   is_public=False, size=13,
                                   properties={'type': 'kernel'})

        data = self.client.get_image(UUID1)

        for k, v in fixture.items():
            el = data[k]
            self.assertEqual(v, data[k],
                             "Failed v != data[k] where v = %(v)s and "
                             "k = %(k)s and data[k] = %(el)s" % {'v': v,
                                                                 'k': k,
                                                                 'el': el})

    def test_get_image_non_existing(self):
        """Tests that NotFound is raised when getting a non-existing image"""
        self.assertRaises(exception.NotFound,
                          self.client.get_image,
                          _gen_uuid())

    def test_add_image_basic(self):
        """Tests that we can add image metadata and returns the new id"""
        fixture = self.get_fixture()

        new_image = self.client.add_image(fixture)

        # Test all other attributes set
        data = self.client.get_image(new_image['id'])

        for k, v in fixture.items():
            self.assertEqual(v, data[k])

        # Test status was updated properly
        self.assertIn('status', data.keys())
        self.assertEqual('active', data['status'])

    def test_add_image_with_properties(self):
        """Tests that we can add image metadata with properties"""
        fixture = self.get_fixture(location="file:///tmp/glance-tests/2",
                                   properties={'distro': 'Ubuntu 10.04 LTS'})

        new_image = self.client.add_image(fixture)

        del fixture['location']
        for k, v in fixture.items():
            self.assertEqual(v, new_image[k])

        # Test status was updated properly
        self.assertIn('status', new_image.keys())
        self.assertEqual('active', new_image['status'])

    def test_add_image_with_location_data(self):
        """Tests that we can add image metadata with properties"""
        location = "file:///tmp/glance-tests/2"
        loc_meta = {'key': 'value'}
        fixture = self.get_fixture(location_data=[{'url': location,
                                                   'metadata': loc_meta,
                                                   'status': 'active'}],
                                   properties={'distro': 'Ubuntu 10.04 LTS'})

        new_image = self.client.add_image(fixture)

        self.assertEqual(location, new_image['location'])
        self.assertEqual(location, new_image['location_data'][0]['url'])
        self.assertEqual(loc_meta, new_image['location_data'][0]['metadata'])

    def test_add_image_with_location_data_with_encryption(self):
        """Tests that we can add image metadata with properties and
        enable encryption.
        """
        self.client.metadata_encryption_key = '1234567890123456'

        location = "file:///tmp/glance-tests/%d"
        loc_meta = {'key': 'value'}
        fixture = {'name': 'fake public image',
                   'is_public': True,
                   'disk_format': 'vmdk',
                   'container_format': 'ovf',
                   'size': 19,
                   'location_data': [{'url': location % 1,
                                      'metadata': loc_meta,
                                      'status': 'active'},
                                     {'url': location % 2,
                                      'metadata': {},
                                      'status': 'active'}],
                   'properties': {'distro': 'Ubuntu 10.04 LTS'}}

        new_image = self.client.add_image(fixture)

        self.assertEqual(location % 1, new_image['location'])
        self.assertEqual(2, len(new_image['location_data']))
        self.assertEqual(location % 1, new_image['location_data'][0]['url'])
        self.assertEqual(loc_meta, new_image['location_data'][0]['metadata'])
        self.assertEqual(location % 2, new_image['location_data'][1]['url'])
        self.assertEqual({}, new_image['location_data'][1]['metadata'])

        self.client.metadata_encryption_key = None

    def test_add_image_already_exists(self):
        """Tests proper exception is raised if image with ID already exists"""
        fixture = self.get_fixture(id=UUID2,
                                   location="file:///tmp/glance-tests/2")

        self.assertRaises(exception.Duplicate,
                          self.client.add_image,
                          fixture)

    def test_add_image_with_bad_status(self):
        """Tests proper exception is raised if a bad status is set"""
        fixture = self.get_fixture(status='bad status',
                                   location="file:///tmp/glance-tests/2")

        self.assertRaises(exception.Invalid,
                          self.client.add_image,
                          fixture)

    def test_update_image(self):
        """Tests that the /images PUT registry API updates the image"""
        fixture = {'name': 'fake public image #2',
                   'disk_format': 'vmdk'}

        self.assertTrue(self.client.update_image(UUID2, fixture))

        # Test all other attributes set
        data = self.client.get_image(UUID2)

        for k, v in fixture.items():
            self.assertEqual(v, data[k])

    def test_update_image_not_existing(self):
        """Tests non existing image update doesn't work"""
        fixture = self.get_fixture(status='bad status')

        self.assertRaises(exception.NotFound,
                          self.client.update_image,
                          _gen_uuid(),
                          fixture)

    def test_delete_image(self):
        """Tests that image metadata is deleted properly"""
        # Grab the original number of images
        orig_num_images = len(self.client.get_images())

        # Delete image #2
        image = self.FIXTURES[1]
        deleted_image = self.client.delete_image(image['id'])
        self.assertTrue(deleted_image)
        self.assertEqual(image['id'], deleted_image['id'])
        self.assertTrue(deleted_image['deleted'])
        self.assertTrue(deleted_image['deleted_at'])

        # Verify one less image
        new_num_images = len(self.client.get_images())

        self.assertEqual(orig_num_images - 1, new_num_images)

    def test_delete_image_not_existing(self):
        """Check that one cannot delete non-existing image."""
        self.assertRaises(exception.NotFound,
                          self.client.delete_image,
                          _gen_uuid())

    def test_get_image_members(self):
        """Test getting image members."""
        memb_list = self.client.get_image_members(UUID2)
        num_members = len(memb_list)
        self.assertEqual(0, num_members)

    def test_get_image_members_not_existing(self):
        """Test getting non-existent image members."""
        self.assertRaises(exception.NotFound,
                          self.client.get_image_members,
                          _gen_uuid())

    def test_get_member_images(self):
        """Test getting member images."""
        memb_list = self.client.get_member_images('pattieblack')
        num_members = len(memb_list)
        self.assertEqual(0, num_members)

    def test_add_replace_members(self):
        """Test replacing image members."""
        self.assertTrue(self.client.add_member(UUID2, 'pattieblack'))
        self.assertTrue(self.client.replace_members(UUID2,
                                                    dict(member_id='pattie'
                                                                   'black2')))

    def test_add_delete_member(self):
        """Tests deleting image members"""
        self.client.add_member(UUID2, 'pattieblack')
        self.assertTrue(self.client.delete_member(UUID2, 'pattieblack'))


class TestBaseClient(testtools.TestCase):

    """
    Test proper actions made for both valid and invalid requests
    against a Registry service
    """
    def test_connect_kwargs_default_values(self):
        actual = test_client.BaseClient('127.0.0.1').get_connect_kwargs()
        self.assertEqual({'timeout': None}, actual)

    def test_connect_kwargs(self):
        base_client = test_client.BaseClient(
            host='127.0.0.1', port=80, timeout=1, use_ssl=True)
        actual = base_client.get_connect_kwargs()
        expected = {'insecure': False,
                    'key_file': None,
                    'cert_file': None,
                    'timeout': 1}
        for k in expected.keys():
            self.assertEqual(expected[k], actual[k])


class TestRegistryV1ClientApi(base.IsolatedUnitTest):

    def setUp(self):
        """Establish a clean test environment."""
        super(TestRegistryV1ClientApi, self).setUp()
        self.context = context.RequestContext()
        reload_module(rapi)

    def tearDown(self):
        """Clear the test environment."""
        super(TestRegistryV1ClientApi, self).tearDown()

    def test_get_registry_client(self):
        actual_client = rapi.get_registry_client(self.context)
        self.assertIsNone(actual_client.identity_headers)

    def test_get_registry_client_with_identity_headers(self):
        self.config(send_identity_headers=True)
        expected_identity_headers = {
            'X-User-Id': '',
            'X-Tenant-Id': '',
            'X-Roles': ','.join(self.context.roles),
            'X-Identity-Status': 'Confirmed',
            'X-Service-Catalog': 'null',
        }
        actual_client = rapi.get_registry_client(self.context)
        self.assertEqual(expected_identity_headers,
                         actual_client.identity_headers)

    def test_configure_registry_client_not_using_use_user_token(self):
        self.config(use_user_token=False)
        with patch.object(rapi, 'configure_registry_admin_creds') as mock_rapi:
            rapi.configure_registry_client()
            mock_rapi.assert_called_once_with()

    def _get_fake_config_creds(self, auth_url='auth_url', strategy='keystone'):
        return {
            'user': 'user',
            'password': 'password',
            'username': 'user',
            'tenant': 'tenant',
            'auth_url': auth_url,
            'strategy': strategy,
            'region': 'region'
        }

    def test_configure_registry_admin_creds(self):
        expected = self._get_fake_config_creds(auth_url=None,
                                               strategy='configured_strategy')
        self.config(admin_user=expected['user'])
        self.config(admin_password=expected['password'])
        self.config(admin_tenant_name=expected['tenant'])
        self.config(auth_strategy=expected['strategy'])
        self.config(auth_region=expected['region'])
        self.stubs.Set(os, 'getenv', lambda x: None)

        self.assertIsNone(rapi._CLIENT_CREDS)
        rapi.configure_registry_admin_creds()
        self.assertEqual(expected, rapi._CLIENT_CREDS)

    def test_configure_registry_admin_creds_with_auth_url(self):
        expected = self._get_fake_config_creds()
        self.config(admin_user=expected['user'])
        self.config(admin_password=expected['password'])
        self.config(admin_tenant_name=expected['tenant'])
        self.config(auth_url=expected['auth_url'])
        self.config(auth_strategy='test_strategy')
        self.config(auth_region=expected['region'])

        self.assertIsNone(rapi._CLIENT_CREDS)
        rapi.configure_registry_admin_creds()
        self.assertEqual(expected, rapi._CLIENT_CREDS)


class FakeResponse(object):
    status = 202

    def getheader(*args, **kwargs):
        return None


class TestRegistryV1ClientRequests(base.IsolatedUnitTest):

    def setUp(self):
        super(TestRegistryV1ClientRequests, self).setUp()

    def tearDown(self):
        super(TestRegistryV1ClientRequests, self).tearDown()

    def test_do_request_with_identity_headers(self):
        identity_headers = {'foo': 'bar'}
        self.client = rclient.RegistryClient("0.0.0.0",
                                             identity_headers=identity_headers)

        with patch.object(test_client.BaseClient, 'do_request',
                          return_value=FakeResponse()) as mock_do_request:
            self.client.do_request("GET", "/images")
            mock_do_request.assert_called_once_with("GET", "/images",
                                                    headers=identity_headers)

    def test_do_request(self):
        self.client = rclient.RegistryClient("0.0.0.0")

        with patch.object(test_client.BaseClient, 'do_request',
                          return_value=FakeResponse()) as mock_do_request:
            self.client.do_request("GET", "/images")
            mock_do_request.assert_called_once_with("GET", "/images",
                                                    headers={})

    def test_registry_invalid_token_exception_handling(self):
        self.image_controller = acontroller()
        request = webob.Request.blank('/images')
        request.method = 'GET'
        request.context = context.RequestContext()

        with patch.object(rapi, 'get_images_detail') as mock_detail:
            mock_detail.side_effect = exception.NotAuthenticated()
            self.assertRaises(webob.exc.HTTPUnauthorized,
                              self.image_controller.detail, request)



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six

from glance.common import exception


class Showlevel(object):
    # None - do not show additional properties and blobs with locations;
    # Basic - show all artifact fields except dependencies;
    # Direct - show all artifact fields with only direct dependencies;
    # Transitive - show all artifact fields with all of dependencies.
    NONE = 0
    BASIC = 1
    DIRECT = 2
    TRANSITIVE = 3

    _level_map = {'none': NONE, 'basic': BASIC, 'direct': DIRECT,
                  'transitive': TRANSITIVE}
    _inverted_level_map = {v: k for k, v in six.iteritems(_level_map)}

    @staticmethod
    def to_str(n):
        try:
            return Showlevel._inverted_level_map[n]
        except KeyError:
            raise exception.ArtifactUnsupportedShowLevel()

    @staticmethod
    def from_str(str_value):
        try:
            return Showlevel._level_map[str_value]
        except KeyError:
            raise exception.ArtifactUnsupportedShowLevel()



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import functools

import mock
import six

import glance.common.exception as exc
from glance.common.glare import declarative
import glance.common.glare.definitions as defs
from glance.common.glare import serialization
import glance.tests.utils as test_utils


BASE = declarative.get_declarative_base()


class TestDeclarativeProperties(test_utils.BaseTestCase):
    def test_artifact_type_properties(self):
        class SomeTypeWithNoExplicitName(BASE):
            some_attr = declarative.AttributeDefinition()

        class InheritedType(SomeTypeWithNoExplicitName):
            __type_version__ = '1.0'
            __type_name__ = 'ExplicitName'
            __type_description__ = 'Type description'
            __type_display_name__ = 'EXPLICIT_NAME'
            __endpoint__ = 'some_endpoint'

            some_attr = declarative.AttributeDefinition(display_name='NAME')

        base_type = SomeTypeWithNoExplicitName
        base_instance = SomeTypeWithNoExplicitName()
        self.assertIsNotNone(base_type.metadata)
        self.assertIsNotNone(base_instance.metadata)
        self.assertEqual(base_type.metadata, base_instance.metadata)
        self.assertEqual("SomeTypeWithNoExplicitName",
                         base_type.metadata.type_name)
        self.assertEqual("SomeTypeWithNoExplicitName",
                         base_type.metadata.type_display_name)
        self.assertEqual("1.0", base_type.metadata.type_version)
        self.assertIsNone(base_type.metadata.type_description)
        self.assertEqual('sometypewithnoexplicitname',
                         base_type.metadata.endpoint)

        self.assertIsNone(base_instance.some_attr)
        self.assertIsNotNone(base_type.some_attr)
        self.assertEqual(base_type.some_attr,
                         base_instance.metadata.attributes.all['some_attr'])
        self.assertEqual('some_attr', base_type.some_attr.name)
        self.assertEqual('some_attr', base_type.some_attr.display_name)
        self.assertIsNone(base_type.some_attr.description)

        derived_type = InheritedType
        derived_instance = InheritedType()

        self.assertIsNotNone(derived_type.metadata)
        self.assertIsNotNone(derived_instance.metadata)
        self.assertEqual(derived_type.metadata, derived_instance.metadata)
        self.assertEqual('ExplicitName', derived_type.metadata.type_name)
        self.assertEqual('EXPLICIT_NAME',
                         derived_type.metadata.type_display_name)
        self.assertEqual('1.0', derived_type.metadata.type_version)
        self.assertEqual('Type description',
                         derived_type.metadata.type_description)
        self.assertEqual('some_endpoint', derived_type.metadata.endpoint)
        self.assertIsNone(derived_instance.some_attr)
        self.assertIsNotNone(derived_type.some_attr)
        self.assertEqual(derived_type.some_attr,
                         derived_instance.metadata.attributes.all['some_attr'])
        self.assertEqual('some_attr', derived_type.some_attr.name)
        self.assertEqual('NAME', derived_type.some_attr.display_name)

    def test_wrong_type_definition(self):
        def declare_wrong_type_version():
            class WrongType(BASE):
                __type_version__ = 'abc'  # not a semver

            return WrongType

        def declare_wrong_type_name():
            class WrongType(BASE):
                __type_name__ = 'a' * 256  # too long

            return WrongType

        self.assertRaises(exc.InvalidArtifactTypeDefinition,
                          declare_wrong_type_version)
        self.assertRaises(exc.InvalidArtifactTypeDefinition,
                          declare_wrong_type_name)

    def test_base_declarative_attributes(self):
        class TestType(BASE):
            defaulted = declarative.PropertyDefinition(default=42)
            read_only = declarative.PropertyDefinition(readonly=True)
            required_attr = declarative.PropertyDefinition(required=True)

        e = self.assertRaises(exc.InvalidArtifactPropertyValue, TestType)
        self.assertEqual('required_attr', e.name)
        self.assertIsNone(e.value)
        tt = TestType(required_attr="universe")
        self.assertEqual('universe', tt.required_attr)
        self.assertEqual(42, tt.defaulted)
        self.assertIsNone(tt.read_only)

        tt = TestType(required_attr="universe", defaulted=0, read_only="Hello")
        self.assertEqual(0, tt.defaulted)
        self.assertEqual("Hello", tt.read_only)

        tt.defaulted = 5
        self.assertEqual(5, tt.defaulted)
        tt.required_attr = 'Foo'
        self.assertEqual('Foo', tt.required_attr)

        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'read_only', 'some_val')
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'required_attr', None)

        # no type checks in base AttributeDefinition
        o = object()
        tt.required_attr = o
        self.assertEqual(o, tt.required_attr)

    def test_generic_property(self):
        class TestType(BASE):
            simple_prop = declarative.PropertyDefinition()
            immutable_internal = declarative.PropertyDefinition(mutable=False,
                                                                internal=True)
            prop_with_allowed = declarative.PropertyDefinition(
                allowed_values=["Foo", True, 42])

        class DerivedType(TestType):
            prop_with_allowed = declarative.PropertyDefinition(
                allowed_values=["Foo", True, 42], required=True, default=42)

        tt = TestType()
        self.assertEqual(True,
                         tt.metadata.attributes.all['simple_prop'].mutable)
        self.assertEqual(False,
                         tt.metadata.attributes.all['simple_prop'].internal)
        self.assertEqual(False,
                         tt.metadata.attributes.all[
                             'immutable_internal'].mutable)
        self.assertEqual(True,
                         tt.metadata.attributes.all[
                             'immutable_internal'].internal)
        self.assertIsNone(tt.prop_with_allowed)
        tt = TestType(prop_with_allowed=42)
        self.assertEqual(42, tt.prop_with_allowed)
        tt = TestType(prop_with_allowed=True)
        self.assertEqual(True, tt.prop_with_allowed)
        tt = TestType(prop_with_allowed='Foo')
        self.assertEqual('Foo', tt.prop_with_allowed)

        tt.prop_with_allowed = 42
        self.assertEqual(42, tt.prop_with_allowed)
        tt.prop_with_allowed = 'Foo'
        self.assertEqual('Foo', tt.prop_with_allowed)
        tt.prop_with_allowed = True
        self.assertEqual(True, tt.prop_with_allowed)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr,
                          tt, 'prop_with_allowed', 'bar')
        # ensure that wrong assignment didn't change the value
        self.assertEqual(True, tt.prop_with_allowed)
        self.assertRaises(exc.InvalidArtifactPropertyValue, TestType,
                          prop_with_allowed=False)

        dt = DerivedType()
        self.assertEqual(42, dt.prop_with_allowed)

    def test_default_violates_allowed(self):
        def declare_wrong_type():
            class WrongType(BASE):
                prop = declarative.PropertyDefinition(
                    allowed_values=['foo', 'bar'],
                    default='baz')

            return WrongType

        self.assertRaises(exc.InvalidArtifactTypePropertyDefinition,
                          declare_wrong_type)

    def test_string_property(self):
        class TestType(BASE):
            simple = defs.String()
            with_length = defs.String(max_length=10, min_length=5)
            with_pattern = defs.String(pattern='^\\d+$', default='42')

        tt = TestType()
        tt.simple = 'foo'
        self.assertEqual('foo', tt.simple)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr,
                          tt, 'simple', 42)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr,
                          tt, 'simple', 'x' * 256)
        self.assertRaises(exc.InvalidArtifactPropertyValue, TestType,
                          simple='x' * 256)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr,
                          tt, 'with_length', 'x' * 11)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr,
                          tt, 'with_length', 'x' * 4)
        tt.simple = 'x' * 5
        self.assertEqual('x' * 5, tt.simple)
        tt.simple = 'x' * 10
        self.assertEqual('x' * 10, tt.simple)

        self.assertEqual("42", tt.with_pattern)
        tt.with_pattern = '0'
        self.assertEqual('0', tt.with_pattern)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'with_pattern', 'abc')
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'with_pattern', '.123.')

    def test_binary_object_mutable(self):
        def declare_blob(mutable):
            class BLOB(BASE):
                prop = defs.BinaryObject(mutable=mutable)

            return BLOB

        blob = declare_blob(False)()
        self.assertFalse(blob.metadata.attributes.all['prop'].mutable)
        self.assertRaises(exc.InvalidArtifactTypePropertyDefinition,
                          functools.partial(declare_blob, True))

    def test_default_and_allowed_violates_string_constrains(self):
        def declare_wrong_default():
            class WrongType(BASE):
                prop = defs.String(min_length=4, default='foo')

            return WrongType

        def declare_wrong_allowed():
            class WrongType(BASE):
                prop = defs.String(min_length=4, allowed_values=['foo', 'bar'])

            return WrongType

        self.assertRaises(exc.InvalidArtifactTypePropertyDefinition,
                          declare_wrong_default)
        self.assertRaises(exc.InvalidArtifactTypePropertyDefinition,
                          declare_wrong_allowed)

    def test_integer_property(self):
        class TestType(BASE):
            simple = defs.Integer()
            constrained = defs.Integer(min_value=10, max_value=50)

        tt = TestType()
        self.assertIsNone(tt.simple)
        self.assertIsNone(tt.constrained)

        tt.simple = 0
        tt.constrained = 10
        self.assertEqual(0, tt.simple)
        self.assertEqual(10, tt.constrained)

        tt.constrained = 50
        self.assertEqual(50, tt.constrained)

        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'constrained', 1)
        self.assertEqual(50, tt.constrained)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'constrained', 51)
        self.assertEqual(50, tt.constrained)

        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'simple', '11')
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'simple', 10.5)

    def test_default_and_allowed_violates_int_constrains(self):
        def declare_wrong_default():
            class WrongType(BASE):
                prop = defs.Integer(min_value=4, default=1)

            return WrongType

        def declare_wrong_allowed():
            class WrongType(BASE):
                prop = defs.Integer(min_value=4, max_value=10,
                                    allowed_values=[1, 15])

            return WrongType

        self.assertRaises(exc.InvalidArtifactTypePropertyDefinition,
                          declare_wrong_default)
        self.assertRaises(exc.InvalidArtifactTypePropertyDefinition,
                          declare_wrong_allowed)

    def test_numeric_values(self):
        class TestType(BASE):
            simple = defs.Numeric()
            constrained = defs.Numeric(min_value=3.14, max_value=4.1)

        tt = TestType(simple=0.1, constrained=4)
        self.assertEqual(0.1, tt.simple)
        self.assertEqual(4.0, tt.constrained)

        tt.simple = 1
        self.assertEqual(1, tt.simple)
        tt.constrained = 3.14
        self.assertEqual(3.14, tt.constrained)
        tt.constrained = 4.1
        self.assertEqual(4.1, tt.constrained)

        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'simple', 'qwerty')
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'constrained', 3)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          'constrained', 5)

    def test_default_and_allowed_violates_numeric_constrains(self):
        def declare_wrong_default():
            class WrongType(BASE):
                prop = defs.Numeric(min_value=4.0, default=1.1)

            return WrongType

        def declare_wrong_allowed():
            class WrongType(BASE):
                prop = defs.Numeric(min_value=4.0, max_value=10.0,
                                    allowed_values=[1.0, 15.5])

            return WrongType

        self.assertRaises(exc.InvalidArtifactTypePropertyDefinition,
                          declare_wrong_default)
        self.assertRaises(exc.InvalidArtifactTypePropertyDefinition,
                          declare_wrong_allowed)

    def test_same_item_type_array(self):
        class TestType(BASE):
            simple = defs.Array()
            unique = defs.Array(unique=True)
            simple_with_allowed_values = defs.Array(
                defs.String(allowed_values=["Foo", "Bar"]))
            defaulted = defs.Array(defs.Boolean(), default=[True, False])
            constrained = defs.Array(item_type=defs.Numeric(min_value=0),
                                     min_size=3, max_size=5, unique=True)

        tt = TestType(simple=[])
        self.assertEqual([], tt.simple)
        tt.simple.append("Foo")
        self.assertEqual(["Foo"], tt.simple)
        tt.simple.append("Foo")
        self.assertEqual(["Foo", "Foo"], tt.simple)
        self.assertEqual(2, len(tt.simple))
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.simple.append,
                          42)
        tt.simple.pop(1)
        self.assertEqual(["Foo"], tt.simple)
        del tt.simple[0]
        self.assertEqual(0, len(tt.simple))

        tt.simple_with_allowed_values = ["Foo"]
        tt.simple_with_allowed_values.insert(0, "Bar")
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.simple_with_allowed_values.append, "Baz")

        self.assertEqual([True, False], tt.defaulted)
        tt.defaulted.pop()
        self.assertEqual([True], tt.defaulted)
        tt2 = TestType()
        self.assertEqual([True, False], tt2.defaulted)

        self.assertIsNone(tt.constrained)
        tt.constrained = [10, 5, 4]
        self.assertEqual([10, 5, 4], tt.constrained)
        tt.constrained[1] = 15
        self.assertEqual([10, 15, 4], tt.constrained)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.constrained.__setitem__, 1, -5)
        self.assertEqual([10, 15, 4], tt.constrained)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.constrained.remove, 15)
        self.assertEqual([10, 15, 4], tt.constrained)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.constrained.__delitem__, 1)
        self.assertEqual([10, 15, 4], tt.constrained)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.constrained.append, 15)
        self.assertEqual([10, 15, 4], tt.constrained)

        tt.unique = []
        tt.unique.append("foo")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.unique.append,
                          "foo")

    def test_tuple_style_array(self):
        class TestType(BASE):
            address = defs.Array(
                item_type=[defs.String(20), defs.Integer(min_value=1),
                           defs.Boolean()])

        tt = TestType(address=["Hope Street", 1234, True])
        self.assertEqual("Hope Street", tt.address[0])
        self.assertEqual(1234, tt.address[1])
        self.assertEqual(True, tt.address[2])

        # On Python 3, sort() fails because int (1) and string ("20") are not
        # comparable
        if six.PY2:
            self.assertRaises(exc.InvalidArtifactPropertyValue,
                              tt.address.sort)
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.address.pop, 0)
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.address.pop, 1)
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.address.pop)
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.address.append,
                          "Foo")

    def test_same_item_type_dict(self):
        class TestType(BASE):
            simple_props = defs.Dict()
            constrained_props = defs.Dict(
                properties=defs.Integer(min_value=1, allowed_values=[1, 2]),
                min_properties=2,
                max_properties=3)

        tt = TestType()
        self.assertIsNone(tt.simple_props)
        self.assertIsNone(tt.constrained_props)
        tt.simple_props = {}
        self.assertEqual({}, tt.simple_props)
        tt.simple_props["foo"] = "bar"
        self.assertEqual({"foo": "bar"}, tt.simple_props)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.simple_props.__setitem__, 42, "foo")
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.simple_props.setdefault, "bar", 42)

        tt.constrained_props = {"foo": 1, "bar": 2}
        self.assertEqual({"foo": 1, "bar": 2}, tt.constrained_props)
        tt.constrained_props["baz"] = 1
        self.assertEqual({"foo": 1, "bar": 2, "baz": 1}, tt.constrained_props)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.constrained_props.__setitem__, "foo", 3)
        self.assertEqual(1, tt.constrained_props["foo"])
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.constrained_props.__setitem__, "qux", 2)
        tt.constrained_props.pop("foo")
        self.assertEqual({"bar": 2, "baz": 1}, tt.constrained_props)
        tt.constrained_props['qux'] = 2
        self.assertEqual({"qux": 2, "bar": 2, "baz": 1}, tt.constrained_props)
        tt.constrained_props.popitem()
        dict_copy = tt.constrained_props.copy()
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.constrained_props.popitem)
        self.assertEqual(dict_copy, tt.constrained_props)

    def test_composite_dict(self):
        class TestType(BASE):
            props = defs.Dict(properties={"foo": defs.String(),
                                          "bar": defs.Boolean()})
            fixed = defs.Dict(properties={"name": defs.String(min_length=2),
                                          "age": defs.Integer(min_value=0,
                                                              max_value=99)})

        tt = TestType()
        tt.props = {"foo": "FOO", "bar": False}
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.props.__setitem__, "bar", 123)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.props.__setitem__, "extra", "value")
        tt.fixed = {"name": "Alex", "age": 42}
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.fixed.__setitem__, "age", 120)

    def test_immutables(self):
        class TestType(BASE):
            activated = defs.Boolean(required=True, default=False)
            name = defs.String(mutable=False)

            def __is_mutable__(self):
                return not self.activated

        tt = TestType()
        self.assertEqual(False, tt.activated)
        self.assertIsNone(tt.name)
        tt.name = "Foo"
        self.assertEqual("Foo", tt.name)
        tt.activated = True
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr,
                          tt, "name", "Bar")
        self.assertEqual("Foo", tt.name)
        tt.activated = False
        tt.name = "Bar"
        self.assertEqual("Bar", tt.name)

    def test_readonly_array_dict(self):
        class TestType(BASE):
            arr = defs.Array(readonly=True)
            dict = defs.Dict(readonly=True)

        tt = TestType(arr=["Foo", "Bar"], dict={"qux": "baz"})
        self.assertEqual(["Foo", "Bar"], tt.arr)
        self.assertEqual({"qux": "baz"}, tt.dict)
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.append,
                          "Baz")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.insert,
                          0, "Baz")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.__setitem__,
                          0, "Baz")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.remove,
                          "Foo")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.pop)
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.dict.pop,
                          "qux")
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.dict.__setitem__, "qux", "foo")

    def test_mutable_array_dict(self):
        class TestType(BASE):
            arr = defs.Array(mutable=False)
            dict = defs.Dict(mutable=False)
            activated = defs.Boolean()

            def __is_mutable__(self):
                return not self.activated

        tt = TestType()
        tt.arr = []
        tt.dict = {}
        tt.arr.append("Foo")
        tt.arr.insert(0, "Bar")
        tt.dict["baz"] = "qux"
        tt.activated = True
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.append,
                          "Baz")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.insert,
                          0, "Baz")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.__setitem__,
                          0, "Baz")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.remove,
                          "Foo")
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.pop)
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.dict.pop,
                          "qux")
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.dict.__setitem__, "qux", "foo")

    def test_readonly_as_write_once(self):
        class TestType(BASE):
            prop = defs.String(readonly=True)
            arr = defs.Array(readonly=True)

        tt = TestType()
        self.assertIsNone(tt.prop)
        tt.prop = "Foo"
        self.assertEqual("Foo", tt.prop)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt,
                          "prop", "bar")
        tt2 = TestType()
        self.assertIsNone(tt2.prop)
        tt2.prop = None
        self.assertIsNone(tt2.prop)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt2,
                          "prop", None)
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, tt2,
                          "prop", "foo")
        self.assertIsNone(tt.arr)
        tt.arr = ["foo", "bar"]
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.append,
                          'baz')
        self.assertIsNone(tt2.arr)
        tt2.arr = None
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.arr.append,
                          'baz')


class TestArtifactType(test_utils.BaseTestCase):
    def test_create_artifact(self):
        a = defs.ArtifactType(**get_artifact_fixture())
        self.assertIsNotNone(a)
        self.assertEqual("123", a.id)
        self.assertEqual("ArtifactType", a.type_name)
        self.assertEqual("1.0", a.type_version)
        self.assertEqual("11.2", a.version)
        self.assertEqual("Foo", a.name)
        self.assertEqual("private", a.visibility)
        self.assertEqual("creating", a.state)
        self.assertEqual("my_tenant", a.owner)
        self.assertEqual(a.created_at, a.updated_at)
        self.assertIsNone(a.description)
        self.assertIsNone(a.published_at)
        self.assertIsNone(a.deleted_at)

        self.assertIsNone(a.description)

        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, a, "id",
                          "foo")
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, a,
                          "state", "active")
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, a,
                          "owner", "some other")
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, a,
                          "created_at", datetime.datetime.now())
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, a,
                          "deleted_at", datetime.datetime.now())
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, a,
                          "updated_at", datetime.datetime.now())
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, a,
                          "published_at", datetime.datetime.now())
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, a,
                          "visibility", "wrong")

    def test_dependency_prop(self):
        class DerivedType(defs.ArtifactType):
            depends_on_any = defs.ArtifactReference()
            depends_on_self = defs.ArtifactReference(type_name='DerivedType')
            depends_on_self_version = defs.ArtifactReference(
                type_name='DerivedType',
                type_version='1.0')

        class DerivedTypeV11(DerivedType):
            __type_name__ = 'DerivedType'
            __type_version__ = '1.1'
            depends_on_self_version = defs.ArtifactReference(
                type_name='DerivedType',
                type_version='1.1')

        d1 = DerivedType(**get_artifact_fixture())
        d2 = DerivedTypeV11(**get_artifact_fixture())
        a = defs.ArtifactType(**get_artifact_fixture())
        d1.depends_on_any = a
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, d1,
                          'depends_on_self', a)
        d1.depends_on_self = d2
        d2.depends_on_self = d1
        d1.depends_on_self_version = d1
        d2.depends_on_self_version = d2
        self.assertRaises(exc.InvalidArtifactPropertyValue, setattr, d1,
                          'depends_on_self_version', d2)

    def test_dependency_list(self):
        class FooType(defs.ArtifactType):
            pass

        class BarType(defs.ArtifactType):
            pass

        class TestType(defs.ArtifactType):
            depends_on = defs.ArtifactReferenceList()
            depends_on_self_or_foo = defs.ArtifactReferenceList(
                references=defs.ArtifactReference(['FooType', 'TestType']))

        a = defs.ArtifactType(**get_artifact_fixture(id="1"))
        a_copy = defs.ArtifactType(**get_artifact_fixture(id="1"))
        b = defs.ArtifactType(**get_artifact_fixture(id="2"))

        tt = TestType(**get_artifact_fixture(id="3"))
        foo = FooType(**get_artifact_fixture(id='4'))
        bar = BarType(**get_artifact_fixture(id='4'))

        tt.depends_on.append(a)
        tt.depends_on.append(b)
        self.assertEqual([a, b], tt.depends_on)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.depends_on.append, a)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.depends_on.append, a_copy)

        tt.depends_on_self_or_foo.append(tt)
        tt.depends_on_self_or_foo.append(foo)
        self.assertRaises(exc.InvalidArtifactPropertyValue,
                          tt.depends_on_self_or_foo.append, bar)
        self.assertEqual([tt, foo], tt.depends_on_self_or_foo)

    def test_blob(self):
        class TestType(defs.ArtifactType):
            image_file = defs.BinaryObject(max_file_size=201054,
                                           min_locations=1,
                                           max_locations=5)
            screen_shots = defs.BinaryObjectList(
                objects=defs.BinaryObject(min_file_size=100), min_count=1)

        tt = TestType(**get_artifact_fixture())
        blob = defs.Blob()
        blob.size = 1024
        blob.locations.append("file://some.file.path")
        tt.image_file = blob

        self.assertEqual(1024, tt.image_file.size)
        self.assertEqual(["file://some.file.path"], tt.image_file.locations)

    def test_pre_publish_blob_validation(self):
        class TestType(defs.ArtifactType):
            required_blob = defs.BinaryObject(required=True)
            optional_blob = defs.BinaryObject()

        tt = TestType(**get_artifact_fixture())
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.__pre_publish__)
        tt.required_blob = defs.Blob(size=0)
        tt.__pre_publish__()

    def test_pre_publish_dependency_validation(self):
        class TestType(defs.ArtifactType):
            required_dependency = defs.ArtifactReference(required=True)
            optional_dependency = defs.ArtifactReference()

        tt = TestType(**get_artifact_fixture())
        self.assertRaises(exc.InvalidArtifactPropertyValue, tt.__pre_publish__)
        tt.required_dependency = defs.ArtifactType(**get_artifact_fixture())
        tt.__pre_publish__()

    def test_default_value_of_immutable_field_in_active_state(self):
        class TestType(defs.ArtifactType):
            foo = defs.String(default='Bar', mutable=False)
        tt = TestType(**get_artifact_fixture(state='active'))
        self.assertEqual('Bar', tt.foo)


class SerTestType(defs.ArtifactType):
    some_string = defs.String()
    some_text = defs.Text()
    some_version = defs.SemVerString()
    some_int = defs.Integer()
    some_numeric = defs.Numeric()
    some_bool = defs.Boolean()
    some_array = defs.Array()
    another_array = defs.Array(
        item_type=[defs.Integer(), defs.Numeric(), defs.Boolean()])
    some_dict = defs.Dict()
    another_dict = defs.Dict(
        properties={'foo': defs.Integer(), 'bar': defs.Boolean()})
    some_ref = defs.ArtifactReference()
    some_ref_list = defs.ArtifactReferenceList()
    some_blob = defs.BinaryObject()
    some_blob_list = defs.BinaryObjectList()


class TestSerialization(test_utils.BaseTestCase):
    def test_serialization_to_db(self):
        ref1 = defs.ArtifactType(**get_artifact_fixture(id="1"))
        ref2 = defs.ArtifactType(**get_artifact_fixture(id="2"))
        ref3 = defs.ArtifactType(**get_artifact_fixture(id="3"))

        blob1 = defs.Blob(size=100, locations=['http://example.com/blob1'],
                          item_key='some_key', checksum='abc')
        blob2 = defs.Blob(size=200, locations=['http://example.com/blob2'],
                          item_key='another_key', checksum='fff')
        blob3 = defs.Blob(size=300, locations=['http://example.com/blob3'],
                          item_key='third_key', checksum='123')

        fixture = get_artifact_fixture()
        tt = SerTestType(**fixture)
        tt.some_string = 'bar'
        tt.some_text = 'bazz'
        tt.some_version = '11.22.33-beta'
        tt.some_int = 50
        tt.some_numeric = 10.341
        tt.some_bool = True
        tt.some_array = ['q', 'w', 'e', 'r', 't', 'y']
        tt.another_array = [1, 1.2, False]
        tt.some_dict = {'foobar': "FOOBAR", 'baz': "QUX"}
        tt.another_dict = {'foo': 1, 'bar': True}
        tt.some_ref = ref1
        tt.some_ref_list = [ref2, ref3]
        tt.some_blob = blob1
        tt.some_blob_list = [blob2, blob3]

        results = serialization.serialize_for_db(tt)
        expected = fixture
        expected['type_name'] = 'SerTestType'
        expected['type_version'] = '1.0'
        expected['properties'] = {
            'some_string': {
                'type': 'string',
                'value': 'bar'
            },
            'some_text': {
                'type': 'text',
                'value': 'bazz'
            },
            'some_version': {
                'type': 'string',
                'value': '11.22.33-beta'
            },
            'some_int': {
                'type': 'int',
                'value': 50
            },
            'some_numeric': {
                'type': 'numeric',
                'value': 10.341
            },
            'some_bool': {
                'type': 'bool',
                'value': True
            },
            'some_array': {
                'type': 'array',
                'value': [
                    {
                        'type': 'string',
                        'value': 'q'
                    },
                    {
                        'type': 'string',
                        'value': 'w'
                    },
                    {
                        'type': 'string',
                        'value': 'e'
                    },
                    {
                        'type': 'string',
                        'value': 'r'
                    },
                    {
                        'type': 'string',
                        'value': 't'
                    },
                    {
                        'type': 'string',
                        'value': 'y'
                    }
                ]
            },
            'another_array': {
                'type': 'array',
                'value': [
                    {
                        'type': 'int',
                        'value': 1
                    },
                    {
                        'type': 'numeric',
                        'value': 1.2
                    },
                    {
                        'type': 'bool',
                        'value': False
                    }
                ]
            },
            'some_dict.foobar': {
                'type': 'string',
                'value': 'FOOBAR'
            },
            'some_dict.baz': {
                'type': 'string',
                'value': 'QUX'
            },
            'another_dict.foo': {
                'type': 'int',
                'value': 1
            },
            'another_dict.bar': {
                'type': 'bool',
                'value': True
            }
        }
        expected['dependencies'] = {
            'some_ref': ['1'],
            'some_ref_list': ['2', '3']
        }
        expected['blobs'] = {
            'some_blob': [
                {
                    'size': 100,
                    'checksum': 'abc',
                    'item_key': 'some_key',
                    'locations': ['http://example.com/blob1']
                }],
            'some_blob_list': [
                {
                    'size': 200,
                    'checksum': 'fff',
                    'item_key': 'another_key',
                    'locations': ['http://example.com/blob2']
                },
                {
                    'size': 300,
                    'checksum': '123',
                    'item_key': 'third_key',
                    'locations': ['http://example.com/blob3']
                }
            ]
        }

        self.assertEqual(expected, results)

    def test_deserialize_from_db(self):
        ts = datetime.datetime.now()
        db_dict = {
            "type_name": 'SerTestType',
            "type_version": '1.0',
            "id": "123",
            "version": "11.2",
            "description": None,
            "name": "Foo",
            "visibility": "private",
            "state": "creating",
            "owner": "my_tenant",
            "created_at": ts,
            "updated_at": ts,
            "deleted_at": None,
            "published_at": None,
            "tags": ["test", "fixture"],
            "properties": {
                'some_string': {
                    'type': 'string',
                    'value': 'bar'
                },
                'some_text': {
                    'type': 'text',
                    'value': 'bazz'
                },
                'some_version': {
                    'type': 'string',
                    'value': '11.22.33-beta'
                },
                'some_int': {
                    'type': 'int',
                    'value': 50
                },
                'some_numeric': {
                    'type': 'numeric',
                    'value': 10.341
                },
                'some_bool': {
                    'type': 'bool',
                    'value': True
                },
                'some_array': {
                    'type': 'array',
                    'value': [
                        {
                            'type': 'string',
                            'value': 'q'
                        },
                        {
                            'type': 'string',
                            'value': 'w'
                        },
                        {
                            'type': 'string',
                            'value': 'e'
                        },
                        {
                            'type': 'string',
                            'value': 'r'
                        },
                        {
                            'type': 'string',
                            'value': 't'
                        },
                        {
                            'type': 'string',
                            'value': 'y'
                        }
                    ]
                },
                'another_array': {
                    'type': 'array',
                    'value': [
                        {
                            'type': 'int',
                            'value': 1
                        },
                        {
                            'type': 'numeric',
                            'value': 1.2
                        },
                        {
                            'type': 'bool',
                            'value': False
                        }
                    ]
                },
                'some_dict.foobar': {
                    'type': 'string',
                    'value': 'FOOBAR'
                },
                'some_dict.baz': {
                    'type': 'string',
                    'value': 'QUX'
                },
                'another_dict.foo': {
                    'type': 'int',
                    'value': 1
                },
                'another_dict.bar': {
                    'type': 'bool',
                    'value': True
                }
            },
            'blobs': {
                'some_blob': [
                    {
                        'size': 100,
                        'checksum': 'abc',
                        'item_key': 'some_key',
                        'locations': ['http://example.com/blob1']
                    }],
                'some_blob_list': [
                    {
                        'size': 200,
                        'checksum': 'fff',
                        'item_key': 'another_key',
                        'locations': ['http://example.com/blob2']
                    },
                    {
                        'size': 300,
                        'checksum': '123',
                        'item_key': 'third_key',
                        'locations': ['http://example.com/blob3']
                    }
                ]
            },
            'dependencies': {
                'some_ref': [
                    {
                        "type_name": 'ArtifactType',
                        "type_version": '1.0',
                        "id": "1",
                        "version": "11.2",
                        "description": None,
                        "name": "Foo",
                        "visibility": "private",
                        "state": "creating",
                        "owner": "my_tenant",
                        "created_at": ts,
                        "updated_at": ts,
                        "deleted_at": None,
                        "published_at": None,
                        "tags": ["test", "fixture"],
                        "properties": {},
                        "blobs": {},
                        "dependencies": {}
                    }
                ],
                'some_ref_list': [
                    {
                        "type_name": 'ArtifactType',
                        "type_version": '1.0',
                        "id": "2",
                        "version": "11.2",
                        "description": None,
                        "name": "Foo",
                        "visibility": "private",
                        "state": "creating",
                        "owner": "my_tenant",
                        "created_at": ts,
                        "updated_at": ts,
                        "deleted_at": None,
                        "published_at": None,
                        "tags": ["test", "fixture"],
                        "properties": {},
                        "blobs": {},
                        "dependencies": {}
                    },
                    {
                        "type_name": 'ArtifactType',
                        "type_version": '1.0',
                        "id": "3",
                        "version": "11.2",
                        "description": None,
                        "name": "Foo",
                        "visibility": "private",
                        "state": "creating",
                        "owner": "my_tenant",
                        "created_at": ts,
                        "updated_at": ts,
                        "deleted_at": None,
                        "published_at": None,
                        "tags": ["test", "fixture"],
                        "properties": {},
                        "blobs": {},
                        "dependencies": {}
                    }
                ]
            }
        }
        plugins_dict = {'SerTestType': [SerTestType],
                        'ArtifactType': [defs.ArtifactType]}

        def _retrieve_plugin(name, version):
            return next((p for p in plugins_dict.get(name, [])
                        if version and p.version == version),
                        plugins_dict.get(name, [None])[0])
        plugins = mock.Mock()
        plugins.get_class_by_typename = _retrieve_plugin
        art = serialization.deserialize_from_db(db_dict, plugins)
        self.assertEqual('123', art.id)
        self.assertEqual('11.2', art.version)
        self.assertIsNone(art.description)
        self.assertEqual('Foo', art.name)
        self.assertEqual('private', art.visibility)
        self.assertEqual('private', art.visibility)


def get_artifact_fixture(**kwargs):
    ts = datetime.datetime.now()
    fixture = {
        "id": "123",
        "version": "11.2",
        "description": None,
        "name": "Foo",
        "visibility": "private",
        "state": "creating",
        "owner": "my_tenant",
        "created_at": ts,
        "updated_at": ts,
        "deleted_at": None,
        "published_at": None,
        "tags": ["test", "fixture"]
    }
    fixture.update(kwargs)
    return fixture



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Functional test cases for glance-replicator"""

import sys

from glance.tests import functional
from glance.tests.utils import execute


class TestGlanceReplicator(functional.FunctionalTest):
    """Functional tests for glance-replicator"""

    def test_compare(self):
        # Test for issue: https://bugs.launchpad.net/glance/+bug/1598928
        cmd = ('%s -m glance.cmd.replicator '
               'compare az1:9292 az2:9292 --debug' %
               (sys.executable,))
        exitcode, out, err = execute(cmd, raise_error=False)
        self.assertIn(
            'Request: GET http://az1:9292/v1/images/detail?is_public=None',
            err
        )
